#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from __future__ import annotations
from abc import abstractmethod
from typing import Callable

import matplotlib
import matplotlib.pyplot as plt
import numpy as np

import os
import importlib
import warnings

import pytopomap.plot as plt_fn
import tilupy.notations as notations
import tilupy.utils as utils
import tilupy.plot as plt_tlp
import tilupy.raster


ALLOWED_MODELS = ["shaltop", "lave2D", "saval2D"]
"""Allowed models for result reading."""


RAW_STATES = ["hvert", "h", "ux", "uy"]
"""Raw states at the output of a model.

Implemented states :

    - hvert : Fluid thickness taken vertically
    - h : Fluid thickness taken normal to topography
    - ux : X-component of fluid velocity
    - uy : Y-component of fluid velocity
"""

TEMPORAL_DATA_0D = ["ek", "volume"]
"""Time-varying 0D data.
   
Implemented 0D temporal data :

    - ek : kinetic energy
    - volume : Fluid volume
    
Also combine all the assembly possibilities between :data:`TEMPORAL_DATA_2D` and :data:`NP_OPERATORS` (or :data:`OTHER_OPERATORS`), at each point xy following this format:

`[TEMPORAL_DATA_2D]_[NP/OTHER_OPERATORS]_xy`

For instance with h :
    - h_max_xy: Maximum value of h over the entire surface for each time step.
    - h_min_xy: Minimal value of h over the entire surface for each time step.
    - h_mean_xy: Mean value of h over the entire surface for each time step.
    - h_std_xy: Standard deviation of h over the entire surface for each time step.
    - h_sum_xy: Sum of each value of h at each point of the surface for each time step.
    - h_int_xy: Integrated value of h at each point of the surface for each time step.
"""

TEMPORAL_DATA_1D = []
"""Time-varying 1D data.

Combine all the assembly possibilities between :data:`TEMPORAL_DATA_2D` and :data:`NP_OPERATORS` (or :data:`OTHER_OPERATORS`) and with an axis like this:

`[TEMPORAL_DATA_2D]_[NP/OTHER_OPERATORS]_[x/y]`

For instance with h :
    - h_max_x: For each Y coordinate, maximum value of h integrating the values of all points on the X axis (along the fixed Y axis) and integrating all time steps, giving hmax(y).
    - h_max_y: For each X coordinate, maximum value of h integrating the values of all points on the Y axis (along the fixed X axis) and integrating all time steps, giving hmax(x).
    - h_min_x: For each Y coordinate, minimum value of h integrating the values of all points on the X axis (along the fixed Y axis) and integrating all time steps, giving hmin(y).
    - h_min_y: For each X coordinate, minimum value of h integrating the values of all points on the Y axis (along the fixed X axis) and integrating all time steps, giving hmin(x).
    - h_mean_x: For each Y coordinate, mean value of h integrating the values of all points on the X axis (along the fixed Y axis) and integrating all time steps, giving hmean(y).
    - h_mean_y: For each X coordinate, mean value of h integrating the values of all points on the Y axis (along the fixed X axis) and integrating all time steps, giving hmean(x).
    - h_std_x: For each Y coordinate, standard deviation of h integrating the values of all points on the X axis (along the fixed Y axis) and integrating all time steps, giving hstd(y).
    - h_std_y: For each X coordinate, standard deviation of h integrating the values of all points on the Y axis (along the fixed X axis) and integrating all time steps, giving hstd(x).
    - h_sum_x: For each Y coordinate, sum of each value of h integrating the values of all points on the X axis (along the fixed Y axis) and integrating all time steps, giving hsum(y).
    - h_sum_y: For each X coordinate, sum of each value of h integrating the values of all points on the Y axis (along the fixed X axis) and integrating all time steps, giving hsum(x).
    - h_int_x: For each Y coordinate, integrate each value of h integrating the values of all points on the X axis (along the fixed Y axis) and integrating all time steps, giving hint(y).
    - h_int_y: For each X coordinate, integrate each value of h integrating the values of all points on the Y axis (along the fixed X axis) and integrating all time steps, giving hint(x).
"""

TEMPORAL_DATA_2D = ["hvert", "h", "u", "ux", "uy", "hu", "hu2"]
"""Time-varying 2D data.
   
Implemented 2D temporal data :

    - hvert : Fluid height taken vertically
    - h : Fluid height taken normal to topography
    - u : Fluid velocity
    - ux : X-component of fluid velocity
    - uy : Y-component of fluid velocity
    - hu : Momentum flux
    - hu2 : Convective momentum flux
"""

STATIC_DATA_0D = []
"""Static 0D data."""

STATIC_DATA_1D = []
"""Static 1D data."""

STATIC_DATA_2D = []
"""Static 2D data.

Combine all the assembly possibilities between :data:`TEMPORAL_DATA_2D` and :data:`NP_OPERATORS` (or :data:`OTHER_OPERATORS`) like this:

`[TEMPORAL_DATA_2D]_[NP/OTHER_OPERATORS]`

For instance with h :
    - h_max : Maximum value of h at each point on the map, integrating all the time steps.
    - h_min : Minimum value of h at each point on the map, integrating all the time steps.
    - h_mean : Mean value of h at each point on the map, integrating all the time steps.
    - h_std : Standard deviation of h at each point on the map, integrating all the time steps.
    - h_sum : Sum of h at each point on the map, integrating all the time steps.
    - h_final : Value of h at each point on the map, for the last time step.
    - h_init : Value of h at each point on the map, for the first time step.
    - h_int : Integrated value of h at each point on the map, integrating all the time steps.
"""

TOPO_DATA_2D = ["z", "zinit", "costh"]
"""Data related to topography.

Implemented topographic data :

    - z : Elevation value of topography
    - zinit : Initial elevation value of topography (same as z if the topography doesn't change during the flow)
    - costh : Cosine of the angle between the vertical and the normal to the relief. Factor to transform vertical height (hvert) into normal height (h).
"""

NP_OPERATORS = ["max", "min", "mean", "std", "sum"]
"""Statistical operators.
   
Implemented operators :

    - max : Maximum
    - min : Minimum
    - mean : Mean
    - std : Standard deviation
    - sum : Sum
"""

OTHER_OPERATORS = ["final", "init", "int"]
"""Other operators.

Implemented operators :

    - final : Final value
    - init : Initial value
    - int : Integrated value
"""

TIME_OPERATORS = ["final", "init", "int"]
"""Time-related operators.

Implemented operators :

    - final : Final value
    - init : Initial value
    - int : Integrated value
"""

for stat in NP_OPERATORS + OTHER_OPERATORS:
    for name in TEMPORAL_DATA_2D:
        STATIC_DATA_2D.append(name + "_" + stat)

for stat in NP_OPERATORS + ["int"]:
    for name in TEMPORAL_DATA_2D:
        TEMPORAL_DATA_0D.append(name + "_" + stat + "_xy")

for stat in NP_OPERATORS + ["int"]:
    for name in TEMPORAL_DATA_2D:
        for axis in ["x", "y"]:
            TEMPORAL_DATA_1D.append(name + "_" + stat + "_" + axis)


TEMPORAL_DATA = TEMPORAL_DATA_0D + TEMPORAL_DATA_1D + TEMPORAL_DATA_2D
"""Assembling all temporal data.

:data:`TEMPORAL_DATA_0D` + :data:`TEMPORAL_DATA_1D` + :data:`TEMPORAL_DATA_2D`
"""

STATIC_DATA = STATIC_DATA_0D + STATIC_DATA_1D + STATIC_DATA_2D
"""Assembling all static data.

:data:`STATIC_DATA_0D` + :data:`STATIC_DATA_1D` + :data:`STATIC_DATA_2D`
"""

DATA_NAMES = (
    TEMPORAL_DATA_0D
    + TEMPORAL_DATA_1D
    + TEMPORAL_DATA_2D
    + STATIC_DATA_0D
    + STATIC_DATA_1D
    + STATIC_DATA_2D
)
"""Assembling all data.

:data:`TEMPORAL_DATA` + :data:`STATIC_DATA`
"""


class AbstractResults:
    """Abstract class for :class:`tilupy.read.TemporalResults` and :class:`tilupy.read.StaticResults`.

    Parameters
    ----------
        name : str
            Name of the property.
        d : numpy.ndarray
            Values of the property.
        notation : dict, optional
            Dictionnary of argument for creating an instance of the class :class:`tilupy.notations.Notation`.
            If None use the function :func:`tilupy.notations.get_notation`. By default None.
        kwargs

    Attributes
    ----------
        _name : str
            Name of the property.
        _d : numpy.ndarray
            Values of the property.
        _notation : tilupy.notations.Notation
            Instance of the class :class:`tilupy.notations.Notation`.
    """

    def __init__(self, name: str, d: np.ndarray, notation: dict = None, **kwargs):
        self._name = name
        self._d = d
        if isinstance(notation, dict):
            self._notation = notations.Notation(**notation)
        elif notation is None:
            self._notation = notations.get_notation(name)
        else:
            self._notation = notation
        self.__dict__.update(kwargs)

    @property
    def d(self) -> np.ndarray:
        """Get data values.

        Returns
        -------
        np.ndarray
            Attribute :attr:`_d`.
        """
        return self._d

    @property
    def name(self) -> str:
        """Get data name.

        Returns
        -------
        str
            Attribute :attr:`_name`.
        """
        return self._name

    @property
    def notation(self) -> tilupy.notations.Notation:
        """Get data notation.

        Returns
        -------
        tilupy.notations.Notation
            Attribute :attr:`_notation`.
        """
        return self._notation


class TemporalResults(AbstractResults):
    """Abstract class for time dependent result of simulation.

    Parameters
    ----------
        name : str
            Name of the property.
        d : numpy.ndarray
            Values of the property.
        t : numpy.ndarray
            Time steps, must match the last dimension of :data:`d`.
        notation : dict, optional
            Dictionnary of argument for creating an instance of the class :class:`tilupy.notations.Notation`.
            If None use the function :func:`tilupy.notations.get_notation`. By default None.

    Attributes
    ----------
        _name : str
            Name of the property.
        _d : numpy.ndarray
            Values of the property.
        _notation : tilupy.notations.Notation
            Instance of the class :class:`tilupy.notations.Notation`.
        _t : numpy.ndarray
            Time steps.
    """

    def __init__(self, name: str, d: np.ndarray, t: np.ndarray, notation: dict = None):
        super().__init__(name, d, notation=notation)
        # 1d array with times, matching last dimension of self.d
        self._t = t

    def get_temporal_stat(
        self, stat: str
    ) -> tilupy.read.StaticResults2D | tilupy.read.StaticResults1D:
        """Statistical analysis along temporal dimension.

        Parameters
        ----------
        stat : str
            Statistical operator to apply. Must be implemented in :data:`NP_OPERATORS` or in
            :data:`OTHER_OPERATORS`.

        Returns
        -------
        tilupy.read.StaticResults2D or tilupy.read.StaticResults1D
            Static result object depending on the dimensionality of the data.
        """
        if stat in NP_OPERATORS:
            dnew = getattr(np, stat)(self._d, axis=-1)
        elif stat == "final":
            dnew = self._d[..., -1]
        elif stat == "init":
            dnew = self._d[..., 0]
        elif stat == "int":
            dnew = np.trapezoid(self._d, x=self._t)

        notation = notations.add_operator(self._notation, stat, axis="t")

        if dnew.ndim == 2:
            return StaticResults2D(
                self._name + "_" + stat,
                dnew,
                notation=notation,
                x=self.x,
                y=self.y,
                z=self.z,
            )
        elif dnew.ndim == 1:
            return StaticResults1D(
                self._name + "_" + stat,
                dnew,
                notation=notation,
                coords=self._coords,
                coords_name=self._coords_name,
            )

    @abstractmethod
    def get_spatial_stat(self, stat, axis):
        """Abstract method for statistical analysis along spatial dimension.

        Parameters
        ----------
        stat : str
            Statistical operator to apply. Must be implemented in :data:`NP_OPERATORS` or in
            :data:`OTHER_OPERATORS`.
        axis : str
            Axis where to do the analysis.
        """
        pass

    @abstractmethod
    def plot(*arg, **kwargs):
        """Abstract method to plot the temporal evolution of the results."""
        pass

    @abstractmethod
    def save(*arg, **kwargs):
        """Abstract method to save the temporal results."""
        pass

    @abstractmethod
    def extract_from_time_step(*arg, **kwargs):
        """Abstract method to extract data from specific time steps of a TemporalResults."""
        pass

    @property
    def t(self) -> np.ndarray:
        """Get times.

        Returns
        -------
        numpy.ndarray
            Attribute :attr:`_t`
        """
        return self._t


class TemporalResults0D(TemporalResults):
    """
    Class for simulation results described where the data is one or multiple scalar functions of time.

    Parameters
    ----------
        name : str
            Name of the property.
        d : numpy.ndarray
            Values of the property. The last dimension correspond to time.
            It can be a one dimensionnal Nt array, or a two dimensionnal [Nd, Nt] array,
            where Nt is the legnth of :data:`t`, and Nd correspond to the number of scalar values
            of interest (e.g. X and Y coordinates of the center of mass / front).
        t : numpy.ndarray
            Time steps, must match the last dimension of :data:`d` (size Nt).
        scalar_names : list[str]
            List of length Nd containing the names of the scalar fields (one
            name per row of d)
        notation : dict, optional
            Dictionnary of argument for creating an instance of the class :class:`tilupy.notations.Notation`.
            If None use the function :func:`tilupy.notations.get_notation`. By default None.

    Attributes
    ----------
        _name : str
            Name of the property.
        _d : numpy.ndarray
            Values of the property.
        _notation : tilupy.notations.Notation
            Instance of the class :class:`tilupy.notations.Notation`.
        _t : numpy.ndarray
            Time steps.
        _scalar_names : list[str]
            List of names of the scalar fields.
    """

    def __init__(
        self,
        name: str,
        d: np.ndarray,
        t: np.ndarray,
        scalar_names: list[str] = None,
        notation: dict = None,
    ):
        super().__init__(name, d, t, notation=notation)
        self._scalar_names = scalar_names

    def plot(
        self,
        ax: matplotlib.axes._axes.Axes = None,
        figsize: tuple[float] = None,
        **kwargs,
    ) -> matplotlib.axes._axes.Axes:
        """Plot the temporal evolution of the 0D results.

        Parameters
        ----------
        ax : matplotlib.axes._axes.Axes, optional
            Existing matplotlib window, if None create one. By default None.
        figsize : tuple[float], optional
            Size of the figure, by default None.

        Returns
        -------
        matplotlib.axes._axes.Axes
            The created plot.
        """
        if ax is None:
            fig, ax = plt.subplots(1, 1, figsize=figsize, layout="constrained")

        if isinstance(self._d, np.ndarray):
            data = self._d.T
        else:
            data = self._d

        if "color" not in kwargs and self._scalar_names is None:
            color = "black"
            kwargs["color"] = color

        ax.plot(
            self._t, data, label=self._scalar_names, **kwargs
        )  # Remove label=self._scalar_names

        ax.grid(True, alpha=0.3)
        ax.set_xlim(left=min(self._t), right=max(self._t))
        ax.set_xlabel(notations.get_label("t"))
        ax.set_ylabel(notations.get_label(self._notation))

        return ax

    def save(self):
        """Save the temporal 0D results.

        Raises
        ------
        NotImplementedError
            Not implemented yet.
        """
        raise NotImplementedError(
            "Saving method for :class:`tilupy.read.TemporalResults0D` not implemented yet"
        )

    def get_spatial_stat(self, *arg, **kwargs):
        """Statistical analysis along spatial dimension for 0D results.

        Raises
        ------
        NotImplementedError
            Not implemented because irrelevant.
        """
        raise NotImplementedError(
            "Spatial integration of :class:`tilupy.read.Spatialresults0D` is not implemented because non relevant"
        )

    def extract_from_time_step(
        self,
        time_steps: float | list[float],
    ) -> tilupy.read.StaticResults0D | tilupy.read.TemporalResults0D:
        """Extract data from specific time steps.

        Parameters
        ----------
        time_steps : float | list[float]
            Value of time steps to extract data.

        Returns
        -------
        tilupy.read.StaticResults0D | tilupy.read.TemporalResults0D
            Extracted data, the type depends on the time step.
        """
        if isinstance(time_steps, float) or isinstance(time_steps, int):
            t_index = np.argmin(np.abs(self._t - time_steps))

            return StaticResults0D(
                name=self._name, d=self._d[t_index], notation=self._notation
            )

        elif isinstance(time_steps, list):
            time_steps = np.array(time_steps)

        if isinstance(time_steps, np.ndarray):
            if isinstance(self._t, list):
                self._t = np.array(self._t)
            t_distances = np.abs(self._t[None, :] - time_steps[:, None])
            t_indexes = np.argmin(t_distances, axis=1)

            return TemporalResults0D(
                name=self._name,
                d=self._d[t_indexes],
                t=self._t[t_indexes],
                notation=self._notation,
            )

    @property
    def scalar_names(self) -> list[str]:
        """Get list of names of the scalar fields.

        Returns
        -------
        list[str]
            Attribute :attr:`_scalar_names`.
        """
        return self._scalar_names


class TemporalResults1D(TemporalResults):
    """
    Class for simulation results described by one dimension for space and one dimension for time.

    Parameters
    ----------
        name : str
            Name of the property.
        d : numpy.ndarray
            Values of the property. The last dimension correspond to time.
            It can be a one dimensionnal Nt array, or a two dimensionnal [Nd, Nt] array,
            where Nt is the legnth of :data:`t`, and Nd correspond to the number of scalar values
            of interest (e.g. X and Y coordinates of the center of mass / front).
        t : numpy.ndarray
            Time steps, must match the last dimension of :data:`d` (size Nt).
        coords: numpy.ndarray
            Spatial coordinates.
        coords_name: str
            Spatial coordinates name.
        notation : dict, optional
            Dictionnary of argument for creating an instance of the class :class:`tilupy.notations.Notation`.
            If None use the function :func:`tilupy.notations.get_notation`. By default None.

    Attributes
    ----------
        _name : str
            Name of the property.
        _d : numpy.ndarray
            Values of the property.
        _notation : tilupy.notations.Notation
            Instance of the class :class:`tilupy.notations.Notation`.
        _t : numpy.ndarray
            Time steps.
        _coords: numpy.ndarray
            Spatial coordinates.
        _coords_name: str
            Spatial coordinates name.
    """

    def __init__(
        self,
        name: str,
        d: np.ndarray,
        t: np.ndarray,
        coords: np.ndarray = None,
        coords_name: str = None,
        notation: dict = None,
    ):
        super().__init__(name, d, t, notation=notation)
        # x and y arrays
        self._coords = coords
        self._coords_name = coords_name

    def plot(
        self,
        coords=None,
        plot_type: str = "simple",
        figsize: tuple[float] = None,
        ax: matplotlib.axes._axes.Axes = None,
        linestyles: list[str] = None,
        cmap: str = "viridis",
        highlighted_curve: bool = False,
        **kwargs,
    ) -> matplotlib.axes._axes.Axes:
        """Plot the temporal evolution of the 1D results.

        Parameters
        ----------
        coords : numpy.ndarray, optional
            Specified coordinates, if None uses the coordinates implemented when creating the instance (:attr:`_coords`).
            By default None.
        plot_type: str, optional
            Wanted plot :

                - "simple" : Every curve in the same graph
                - "multiples" : Every curve in separate graph
                - "'shotgather" : Shotgather graph

            By default "simple".
        ax : matplotlib.axes._axes.Axes, optional
            Existing matplotlib window, if None create one. By default None
        linestyles : list[str], optional
            Custom linestyles for each time step. If None, colors and styles are auto-assigned.
            Used only for "simple". By default None.
        cmap : str, optional
            Color map for the ploted curves. Used only for "simple". By default "viridis".
        hightlighted_curved : bool, optional
            Option to display all time steps on each graph of the multiples and
            highlight the curve corresponding to the time step of the subgraph. Used only for "multiples".
            By default False.
        kwargs: dict, optional
            Additional arguments for plot functions.

        Returns
        -------
        matplotlib.axes._axes.Axes
            The created plot.

        Raises
        ------
        TypeError
            If missing coordinates.
        """
        if coords is None:
            coords = self._coords
            if coords is None:
                raise TypeError("coords data missing")

        if ax is None and plot_type != "multiples":
            fig, ax = plt.subplots(1, 1, figsize=figsize, layout="constrained")

        if plot_type == "shotgather":
            xlabel = notations.get_label(self._coords_name, with_unit=True)

            if "colorbar_kwargs" not in kwargs:
                kwargs["colorbar_kwargs"] = dict()
            if "label" not in kwargs["colorbar_kwargs"]:
                clabel = notations.get_label(self._notation)
                kwargs["colorbar_kwargs"]["label"] = clabel

            ax = plt_tlp.plot_shotgather(
                self._coords,
                self._t,
                self._d,
                xlabel=xlabel,
                ylabel=notations.get_label("t"),
                **kwargs,
            )

        if plot_type == "simple":
            if linestyles is None or len(linestyles) != (len(self._t)):
                norm = plt.Normalize(vmin=0, vmax=len(self._t) - 1)
                cmap = plt.get_cmap(cmap)

            for i in range(self._d.shape[1]):
                if linestyles is None or len(linestyles) != (len(self._t)):
                    color = cmap(norm(i)) if self._t[i] != 0 else "red"
                    l_style = "-" if self._t[i] != 0 else (0, (1, 4))
                else:
                    color = "black" if self._t[i] != 0 else "red"
                    l_style = linestyles[i] if self._t[i] != 0 else (0, (1, 4))

                ax.plot(
                    self._coords,
                    self._d[:, i],
                    label=f"t={self._t[i]}s",
                    color=color,
                    linestyle=l_style,
                    **kwargs,
                )

            ax.grid(True, alpha=0.3)
            ax.set_xlim(left=min(self._coords), right=max(self._coords))

            ax.set_xlabel(notations.get_label(self._coords_name))
            ax.set_ylabel(notations.get_label(self._notation))

        if plot_type == "multiples":
            cols_nb = 3
            if len(self._t) < 3:
                cols_nb = len(self._t)

            row_nb = len(self._t) // 3
            if len(self._t) % 3 != 0:
                row_nb += 1

            fig, axes = plt.subplots(
                nrows=row_nb,
                ncols=cols_nb,
                figsize=figsize,
                layout="constrained",
                sharex=True,
                sharey=True,
            )
            axes = axes.flatten()

            for i in range(self._d.shape[1]):
                if highlighted_curve:
                    for j in range(self._d.shape[1]):
                        if i == j:
                            axes[i].plot(
                                self._coords,
                                self._d[:, j],
                                color="black",
                                linewidth=1.5,
                                **kwargs,
                            )
                        else:
                            axes[i].plot(
                                self._coords,
                                self._d[:, j],
                                color="gray",
                                alpha=0.5,
                                linewidth=0.5,
                                **kwargs,
                            )
                else:
                    axes[i].plot(self._coords, self._d[:, i], color="black", **kwargs)

                axes[i].grid(True, alpha=0.3)
                axes[i].set_xlim(left=min(self._coords), right=max(self._coords))

                axes[i].set_xlabel(notations.get_label(self._coords_name))
                axes[i].set_ylabel(notations.get_label(self._notation))

                axes[i].set_title(f"t={self._t[i]}s", loc="left")

            for i in range(len(self._t), len(axes)):
                fig.delaxes(axes[i])

            return axes

        return ax

    def save(self):
        """Save the temporal 1D results.

        Raises
        ------
        NotImplementedError
            Not implemented yet.
        """
        raise NotImplementedError(
            "Saving method for TemporalResults1D not implemented yet"
        )

    def get_spatial_stat(self, stat, **kwargs) -> tilupy.read.TemporalResults0D:
        """Statistical analysis along spatial dimension for 1D results.

        Parameters
        ----------
        stat : str
            Statistical operator to apply. Must be implemented in :data:`NP_OPERATORS` or in
            :data:`OTHER_OPERATORS`.

        Returns
        -------
        tilupy.read.TemporalResults0D
            Instance of :class:`tilupy.read.TemporalResults0D`.
        """
        if stat in NP_OPERATORS:
            dnew = getattr(np, stat)(self._d, axis=0)
        elif stat == "int":
            dd = self._coords[1] - self._coords[0]
            dnew = np.sum(self._d, axis=0) * dd
        notation = notations.add_operator(self._notation, stat, axis=self._coords_name)

        return TemporalResults0D(
            self._name + "_" + stat, dnew, self._t, notation=notation
        )

    def extract_from_time_step(
        self,
        time_steps: float | list[float],
    ) -> tilupy.read.StaticResults1D | tilupy.read.TemporalResults1D:
        """Extract data from specific time steps.

        Parameters
        ----------
        time_steps : float | list[float]
            Value of time steps to extract data.

        Returns
        -------
        tilupy.read.StaticResults1D | tilupy.read.TemporalResults1D
            Extracted data, the type depends on the time step.
        """
        if isinstance(time_steps, float) or isinstance(time_steps, int):
            t_index = np.argmin(np.abs(self._t - time_steps))

            return StaticResults1D(
                name=self._name,
                d=self._d[:, t_index],
                coords=self._coords,
                coords_name=self._coords_name,
                notation=self._notation,
            )

        elif isinstance(time_steps, list):
            time_steps = np.array(time_steps)

        if isinstance(time_steps, np.ndarray):
            if isinstance(self._t, list):
                self._t = np.array(self._t)
            t_distances = np.abs(self._t[None, :] - time_steps[:, None])
            t_indexes = np.argmin(t_distances, axis=1)

            return TemporalResults1D(
                name=self._name,
                d=self._d[:, t_indexes],
                t=self._t[t_indexes],
                coords=self._coords,
                coords_name=self._coords_name,
                notation=self._notation,
            )

    @property
    def coords(self) -> np.ndarray:
        """Get spatial coordinates.

        Returns
        -------
        numpy.ndarray
            Attribute :attr:`_coords`
        """
        return self._coords

    @property
    def coords_name(self) -> str:
        """Get spatial coordinates name.

        Returns
        -------
        str
            Attribute :attr:`_coords_name`
        """
        return self._coords_name


class TemporalResults2D(TemporalResults):
    """
    Class for simulation results described by a two dimensional space and one dimension for time.

    Parameters
    ----------
        name : str
            Name of the property.
        d : numpy.ndarray
            Values of the property. The last dimension correspond to time.
            It can be a one dimensionnal Nt array, or a two dimensionnal [Nd, Nt] array,
            where Nt is the legnth of :data:`t`, and Nd correspond to the number of scalar values
            of interest (e.g. X and Y coordinates of the center of mass / front).
        t : numpy.ndarray
            Time steps, must match the last dimension of :data:`d` (size Nt).
        x : numpy.ndarray
            X coordinate values.
        y : numpy.ndarray
            X coordinate values.
        z : numpy.ndarray
            Elevation values of the surface.
        notation : dict, optional
            Dictionnary of argument for creating an instance of the class :class:`tilupy.notation.Notation`.
            If None use the function :func:`tilupy.notation.get_notation`. By default None.

    Attributes
    ----------
        _name : str
            Name of the property.
        _d : numpy.ndarray
            Values of the property.
        _notation : tilupy.notations.Notation
            Instance of the class :class:`tilupy.notations.Notation`.
        _t : numpy.ndarray
            Time steps.
        _x : numpy.ndarray
            X coordinate values.
        _y : numpy.ndarray
            X coordinate values.
        _z : numpy.ndarray
            Elevation values of the surface.
    """

    def __init__(
        self,
        name: str,
        d: np.ndarray,
        t: np.ndarray,
        x: np.ndarray = None,
        y: np.ndarray = None,
        z: np.ndarray = None,
        notation: dict = None,
    ):
        super().__init__(name, d, t, notation=notation)
        # x and y arrays
        self._x = x
        self._y = y
        # topography
        self._z = z

    def plot(
        self,
        x: np.ndarray = None,
        y: np.ndarray = None,
        z: np.ndarray = None,
        plot_multiples: bool = False,
        file_name: str = None,
        folder_out: str = None,
        figsize: tuple[float] = None,
        dpi: int = None,
        fmt: str = "png",
        sup_plt_fn=None,
        sup_plt_fn_args=None,
        **kwargs,
    ) -> None:
        """Plot the temporal evolution of the 2D results using :func:`pytopomap.plot.plot_maps`.

        Parameters
        ----------
        x : numpy.ndarray, optional
            X coordinate values, if None use :attr:`_x`. By default None.
        y : numpy.ndarray, optional
            Y coordinate values, if None use :attr:`_y`. By default None.
        z : numpy.ndarray, optional
            Elevation values, if None use :attr:`_z`. By default None.
        file_name : str, optional
            Base name for the output image files, by default None.
        folder_out : str, optional
            Path to the output folder. If not provides, figures are not saved. By default None.
        figsize : tuple[float], optional
            Size of the figure, by default None.
        dpi : int, optional
            Resolution for saved figures. Only used if :data:`folder_out` is set. By default None.
        fmt : str, optional
            File format for saving figures, by default "png".
        sup_plt_fn : callable, optional
            A custom function to apply additional plotting on the axes, by default None.
        sup_plt_fn_args : dict, optional
            Arguments to pass to :data:`sup_plt_fn`, by default None.

        Raises
        ------
        TypeError
            If no value for x, y.
        """
        if file_name is None:
            file_name = self._name

        if x is None:
            x = self._x
        if y is None:
            y = self._y
        if z is None:
            z = self._z

        if x is None or y is None:
            raise TypeError("x, y or z data missing")

        if z is None:
            warnings.warn("No topography given.")

        if "colorbar_kwargs" not in kwargs:
            kwargs["colorbar_kwargs"] = dict()
        if "label" not in kwargs["colorbar_kwargs"]:
            clabel = notations.get_label(self._notation)
            kwargs["colorbar_kwargs"]["label"] = clabel

        if plot_multiples:
            if "vmin" not in kwargs:
                kwargs["vmin"] = np.min(self._d)
            if "vmax" not in kwargs:
                kwargs["vmax"] = np.max(self._d)

            cols_nb = 3
            if len(self._t) < 3:
                cols_nb = len(self._t)

            row_nb = len(self._t) // 3
            if len(self._t) % 3 != 0:
                row_nb += 1

            fig, axes = plt.subplots(
                nrows=row_nb,
                ncols=cols_nb,
                figsize=figsize,
                layout="constrained",
                sharex=True,
                sharey=True,
            )
            axes = axes.flatten()

            for i in range(len(self._t)):
                plt_fn.plot_data_on_topo(
                    x=x,
                    y=y,
                    z=z,
                    data=self._d[:, :, i],
                    axe=axes[i],
                    plot_colorbar=False,
                    **kwargs,
                )

                axes[i].set_title(f"t={self._t[i]}s", loc="left")

            for i in range(len(self._t), len(axes)):
                fig.delaxes(axes[i])

            max_val, idx = 0, 0
            for i in range(len(self._t)):
                max_val_t = np.max(axes[i].images[1].get_array())
                if max_val_t > max_val:
                    max_val = max_val_t
                    idx = i
            mappable = axes[idx].images[1]
            fig.colorbar(
                mappable, ax=axes, orientation="vertical", **kwargs["colorbar_kwargs"]
            )

            return axes

        plt_fn.plot_maps(
            x,
            y,
            z,
            self._d,
            self._t,
            file_name=file_name,
            folder_out=folder_out,
            figsize=figsize,
            dpi=dpi,
            fmt=fmt,
            sup_plt_fn=sup_plt_fn,
            sup_plt_fn_args=sup_plt_fn_args,
            **kwargs,
        )

        return None

    def save(
        self,
        folder: str = None,
        file_name: str = None,
        fmt: str = "asc",
        time: str | int = None,
        x: np.ndarray = None,
        y: np.ndarray = None,
        **kwargs,
    ) -> None:
        """Save the temporal 2D results.

        Parameters
        ----------
        folder : str, optional
            Path to the output folder, if None create a folder with :attr:`_name`. By default None.
        file_name : str, optional
            Base name for the output image files, if None use :attr:`_name`. By default None.
        fmt : str, optional
            File format for saving result, by default "asc".
        time : str | int, optional
            Time instants to save the results.

                - If time is string, must be "initial" or "final".
                - If time is int, used as index in :attr:`_t`.
                - If None use every instant in :attr:`_t`.

            By default None.
        x : np.ndarray, optional
            X coordinate values, if None use :attr:`_x`. By default None.
        y : np.ndarray, optional
            Y coordinate values, if None use :attr:`_y`. By default None.

        Returns
        -------
        None

        Raises
        ------
        ValueError
            If no value for x, y.
        """
        if x is None:
            x = self._x
        if y is None:
            y = self._y
        if x is None or y is None:
            raise ValueError("x et y arrays must not be None")

        if file_name is None:
            file_name = self._name

        if folder is not None:
            file_name = os.path.join(folder, file_name)

        if time is not None:
            if isinstance(time, str):
                if time == "final":
                    inds = [self._d.shape[2] - 1]
                elif time == "initial":
                    inds = [0]
            else:
                inds = [np.argmin(time - np.abs(np.array(self._t) - time))]
        else:
            inds = range(len(self._t))

        for i in inds:
            file_out = file_name + "_{:04d}.".format(i) + fmt
            tilupy.raster.write_raster(
                x, y, self._d[:, :, i], file_out, fmt=fmt, **kwargs
            )

    def get_spatial_stat(
        self, stat: str, axis: str | int | tuple[int] = None
    ) -> tilupy.read.TemporalResults0D | tilupy.read.TemporalResults1D:
        """Statistical analysis along spatial dimension for 2D results.

        Parameters
        ----------
        stat : str
            Statistical operator to apply. Must be implemented in :data:`NP_OPERATORS`.
        axis : tuple[int]
            Axis where to do the analysis:

                - If axis is string, replace 'x' by 1, 'y' by 0 and 'xy' by (0, 1).
                - If axis is int, only use 0 or 1.
                - If None use (0, 1). By default None.

        Returns
        -------
        tilupy.read.TemporalResults0D or tilupy.read.TemporalResults1D
            Instance of :class:`tilupy.read.TemporalResults0D` or :class:`tilupy.read.TemporalResults1D`.
        """
        if axis is None:
            axis = (0, 1)

        if isinstance(axis, str):
            axis_str = axis
            if axis == "x":
                axis = 1
            elif axis == "y":
                axis = 0
            elif axis == "xy":
                axis = (0, 1)
        else:
            if axis == 1:
                axis_str = "x"
            elif axis == 0:
                axis_str = "y"
            elif axis == (0, 1):
                axis_str = "xy"

        if stat in NP_OPERATORS:
            dnew = getattr(np, stat)(self._d, axis=axis)
        elif stat == "int":
            dnew = np.sum(self._d, axis=axis)
            if axis == 1:
                dd = self._x[1] - self._x[0]
            elif axis == 0:
                dd = self._y[1] - self._y[0]
            elif axis == (0, 1):
                dd = (self._x[1] - self._x[0]) * (self._y[1] - self._y[0])
            dnew = dnew * dd

        if axis == 1:
            # Needed to get correct orinetation as d[0, 0] is the upper corner
            # of the data, with coordinates x[0], y[-1]
            dnew = np.flip(dnew, axis=0)

        new_name = self._name + "_" + stat + "_" + axis_str
        notation = notations.add_operator(self._notation, stat, axis=axis_str)

        if axis == (0, 1):
            return TemporalResults0D(new_name, dnew, self._t, notation=notation)
        else:
            if axis == 0:
                coords = self._x
                coords_name = "x"
            else:
                coords = self._y
                coords_name = "y"
            return TemporalResults1D(
                new_name,
                dnew,
                self._t,
                coords,
                coords_name=coords_name,
                notation=notation,
            )

    def extract_from_time_step(
        self,
        time_steps: float | list[float],
    ) -> tilupy.read.StaticResults2D | tilupy.read.TemporalResults2D:
        """Extract data from specific time steps.

        Parameters
        ----------
        time_steps : float | list[float]
            Value of time steps to extract data.

        Returns
        -------
        tilupy.read.StaticResults2D | tilupy.read.TemporalResults2D
            Extracted data, the type depends on the time step.
        """
        if isinstance(time_steps, float) or isinstance(time_steps, int):
            t_index = np.argmin(np.abs(self._t - time_steps))

            return StaticResults2D(
                name=self._name,
                d=self._d[:, :, t_index],
                x=self._x,
                y=self._y,
                z=self._z,
                notation=self._notation,
            )

        elif isinstance(time_steps, list):
            time_steps = np.array(time_steps)

        if isinstance(time_steps, np.ndarray):
            if isinstance(self._t, list):
                self._t = np.array(self._t)
            t_distances = np.abs(self._t[None, :] - time_steps[:, None])
            t_indexes = np.argmin(t_distances, axis=1)

            return TemporalResults2D(
                name=self._name,
                d=self._d[:, :, t_indexes],
                t=self._t[t_indexes],
                x=self._x,
                y=self._y,
                z=self._z,
                notation=self._notation,
            )

    @property
    def x(self) -> np.ndarray:
        """Get X coordinates.

        Returns
        -------
        numpy.ndarray
            Attribute :attr:`_x`.
        """
        return self._x

    @property
    def y(self) -> np.ndarray:
        """Get Y coordinates.

        Returns
        -------
        numpy.ndarray
            Attribute :attr:`_y`.
        """
        return self._y

    @property
    def z(self) -> np.ndarray:
        """Get elevations values.

        Returns
        -------
        numpy.ndarray
            Attribute :attr:`_z`.
        """
        return self._z


class StaticResults(AbstractResults):
    """Abstract class for result of simulation without time dependence.

    Parameters
    ----------
        name : str
            Name of the property.
        d : numpy.ndarray
            Values of the property.
        notation : dict, optional
            Dictionnary of argument for creating an instance of the class :class:`tilupy.notations.Notation`.
            If None use the function :func:`tilupy.notations.get_notation`. By default None.

    Attributes
    ----------
        _name : str
            Name of the property.
        _d : numpy.ndarray
            Values of the property.
        _notation : tilupy.notations.Notation
            Instance of the class :class:`tilupy.notations.Notation`.
    """

    def __init__(self, name: str, d: np.ndarray, notation: dict = None):
        super().__init__(name, d, notation=notation)
        # x and y arrays

    @abstractmethod
    def plot(self):
        """Abstract method to plot the results."""
        pass

    @abstractmethod
    def save(self):
        """Abstract method to save the results."""
        pass


class StaticResults0D(StaticResults):
    """
    Class for simulation results described where the data is one or multiple scalar.

    Parameters
    ----------
        name : str
            Name of the property.
        d : numpy.ndarray
            Values of the property.
        notation : dict, optional
            Dictionnary of argument for creating an instance of the class :class:`tilupy.notations.Notation`.
            If None use the function :func:`tilupy.notations.get_notation`. By default None.

    Attributes
    ----------
        _name : str
            Name of the property.
        _d : numpy.ndarray
            Values of the property.
        _notation : tilupy.notations.Notation
            Instance of the class :class:`tilupy.notations.Notation`.
    """

    def __init__(self, name: str, d: np.ndarray, notation: dict = None):
        super().__init__(name, d, notation=notation)


class StaticResults1D(StaticResults):
    """
    Class for simulation results described by one dimension for space.

    Parameters
    ----------
        name : str
            Name of the property.
        d : numpy.ndarray
            Values of the property.
        coords: numpy.ndarray
            Spatial coordinates.
        coords_name: str
            Spatial coordinates name.
        notation : dict, optional
            Dictionnary of argument for creating an instance of the class :class:`tilupy.notations.Notation`.
            If None use the function :func:`tilupy.notations.get_notation`. By default None.

    Attributes
    ----------
        _name : str
            Name of the property.
        _d : numpy.ndarray
            Values of the property.
        _notation : tilupy.notations.Notation
            Instance of the class :class:`tilupy.notations.Notation`.
        _coords: numpy.ndarray
            Spatial coordinates.
        _coords_name: str
            Spatial coordinates name.
    """

    def __init__(
        self,
        name: str,
        d: np.ndarray,
        coords: np.ndarray = None,
        coords_name: list[str] = None,
        notation: dict = None,
    ):
        super().__init__(name, d, notation=notation)
        # x and y arrays
        self._coords = coords
        self._coords_name = coords_name

    def plot(
        self, ax: matplotlib.axes._axes.Axes = None, **kwargs
    ) -> matplotlib.axes._axes.Axes:
        """Plot the 1D results.

        Parameters
        ----------
        ax : matplotlib.axes._axes.Axes, optional
            Existing matplotlib window, if None create one. By default None.

        Returns
        -------
        matplotlib.axes._axes.Axes
            The created plot.
        """
        if ax is None:
            fig, ax = plt.subplots(1, 1, layout="constrained")

        if isinstance(self._d, np.ndarray):
            data = self._d.T
        else:
            data = self._d

        if "color" not in kwargs:
            color = "black"
            kwargs["color"] = color

        ax.plot(self._coords, data, **kwargs)

        ax.grid(True, alpha=0.3)
        ax.set_xlim(left=min(self._coords), right=max(self._coords))
        ax.set_xlabel(notations.get_label(self._coords_name))
        ax.set_ylabel(notations.get_label(self._notation))

        return ax

    @property
    def coords(self) -> np.ndarray:
        """Get spatial coordinates.

        Returns
        -------
        numpy.ndarray
            Attribute :attr:`_coords`
        """
        return self._coords

    @property
    def coords_name(self) -> str:
        """Get spatial coordinates name.

        Returns
        -------
        str
            Attribute :attr:`_coords_name`
        """
        return self._coords_name


class StaticResults2D(StaticResults):
    """
    Class for simulation results described by a two dimensional space result.
    Inherits from StaticResults.

    Parameters
    ----------
        name : str
            Name of the property.
        d : numpy.ndarray
            Values of the property.
        x : numpy.ndarray
            X coordinate values.
        y : numpy.ndarray
            X coordinate values.
        z : numpy.ndarray
            Elevation values of the surface.
        notation : dict, optional
            Dictionnary of argument for creating an instance of the class :class:`tilupy.notations.Notation`.
            If None use the function :func:`tilupy.notations.get_notation`. By default None.

    Attributes
    ----------
        _name : str
            Name of the property.
        _d : numpy.ndarray
            Values of the property.
        _notation : tilupy.notations.Notation
            Instance of the class :class:`tilupy.notations.Notation`.
        _x : numpy.ndarray
            X coordinate values.
        _y : numpy.ndarray
            X coordinate values.
        _z : numpy.ndarray
            Elevation values of the surface.
    """

    def __init__(
        self,
        name: str,
        d: np.ndarray,
        x: np.ndarray = None,
        y: np.ndarray = None,
        z: np.ndarray = None,
        notation: dict = None,
    ):
        super().__init__(name, d, notation=notation)
        # x and y arrays
        self._x = x
        self._y = y
        # topography
        self._z = z

    def plot(
        self,
        figsize: tuple[float] = None,
        x: np.ndarray = None,
        y: np.ndarray = None,
        z: np.ndarray = None,
        sup_plt_fn: Callable = None,
        sup_plt_fn_args: dict = None,
        ax: matplotlib.axes._axes.Axes = None,
        **kwargs,
    ) -> matplotlib.axes._axes.Axes:
        """Plot the 2D results using :func:`pytopomap.plot.plot_data_on_topo`.

        Parameters
        ----------
        figsize : tuple[float], optional
            Size of the figure, by default None
        x : numpy.ndarray, optional
            X coordinate values, if None use :attr:`_x`. By default None.
        y : numpy.ndarray, optional
            Y coordinate values, if None use :attr:`_y`. By default None.
        z : numpy.ndarray, optional
            Elevation values, if None use :attr:`_z`. By default None.
        sup_plt_fn : callable, optional
            A custom function to apply additional plotting on the axes, by default None.
        sup_plt_fn_args : dict, optional
            Arguments to pass to :data:`sup_plt_fn`, by default None.
        ax : matplotlib.axes._axes.Axes, optional
            Existing matplotlib window, if None create one. By default None
        kwargs
            Additional arguments to pass to :func:`pytopomap.plot.plot_data_on_topo`.

        Returns
        -------
        matplotlib.axes._axes.Axes
            The created plot.

        Raises
        ------
        TypeError
            If no value for x, y.
        """
        if ax is None:
            fig, ax = plt.subplots(1, 1, figsize=figsize, layout="constrained")

        if x is None:
            x = self._x
        if y is None:
            y = self._y
        if z is None:
            z = self._z

        if x is None or y is None or z is None:
            raise TypeError("x, y or z data missing")

        if "colorbar_kwargs" not in kwargs:
            kwargs["colorbar_kwargs"] = dict()
        if "label" not in kwargs["colorbar_kwargs"]:
            clabel = notations.get_label(self._notation)
            kwargs["colorbar_kwargs"]["label"] = clabel

        ax = plt_fn.plot_data_on_topo(
            x, y, z, self._d, axe=ax, figsize=figsize, **kwargs
        )
        if sup_plt_fn is not None:
            if sup_plt_fn_args is None:
                sup_plt_fn_args = dict()
            sup_plt_fn(ax, **sup_plt_fn_args)

        return ax

    def save(
        self,
        folder: str = None,
        file_name: str = None,
        fmt: str = "txt",
        x: np.ndarray = None,
        y: np.ndarray = None,
        **kwargs,
    ) -> None:
        """Save the 2D results.

        Parameters
        ----------
        folder : str, optional
            Path to the output folder, if None create a folder with :attr:`_name`. By default None.
        file_name : str, optional
            Base name for the output image files, if None use :attr:`_name`. By default None.
        fmt : str, optional
            File format for saving result, by default "txt".
        x : np.ndarray, optional
            X coordinate values, if None use :attr:`_x`. By default None.
        y : np.ndarray, optional
            Y coordinate values, if None use :attr:`_y`. By default None.

        Raises
        ------
        ValueError
            If no value for x, y.
        """
        if x is None:
            x = self._x
        if y is None:
            y = self._y

        if x is None or y is None:
            raise ValueError("x et y arrays must not be None")

        if file_name is None:
            file_name = self._name + "." + fmt

        if folder is not None:
            file_name = os.path.join(folder, file_name)

        tilupy.raster.write_raster(x, y, self._d, file_name, fmt=fmt, **kwargs)

    def get_spatial_stat(
        self, stat: str, axis=None
    ) -> tilupy.read.StaticResults0D | tilupy.read.StaticResults1D:
        """Statistical analysis along spatial dimension for 2D results.

        Parameters
        ----------
        stat : str
            Statistical operator to apply. Must be implemented in :data:`NP_OPERATORS`.
        axis : tuple[int]
            Axis where to do the analysis:

                - If axis is string, replace 'x' by 1, 'y' by 0 and 'xy' by (0, 1).
                - If axis is int, only use 0 or 1.
                - If None use (0, 1). By default None.

        Returns
        -------
        tilupy.read.StaticResults0D or tilupy.read.StaticResults1D
            Instance of :class:`tilupy.read.StaticResults0D` or :class:`tilupy.read.StaticResults1D`.

        """
        if axis is None:
            axis = (0, 1)

        if isinstance(axis, str):
            axis_str = axis
            if axis == "x":
                axis = 1
            elif axis == "y":
                axis = 0
            elif axis == "xy":
                axis = (0, 1)
        else:
            if axis == 1:
                axis_str = "x"
            elif axis == 0:
                axis_str = "y"
            elif axis == (0, 1):
                axis_str = "xy"

        if stat in NP_OPERATORS:
            dnew = getattr(np, stat)(self._d, axis=axis)
        elif stat == "int":
            dnew = np.sum(self._d, axis=axis)
            if axis == 1:
                dd = self._x[1] - self._x[0]
            elif axis == 0:
                dd = self._y[1] - self._y[0]
            elif axis == (0, 1):
                dd = (self._x[1] - self._x[0]) * (self._y[1] - self._y[0])
            dnew = dnew * dd

        if axis == 1:
            # Needed to get correct orinetation as d[0, 0] is the upper corner
            # of the data, with coordinates x[0], y[-1]
            dnew = np.flip(dnew, axis=0)

        new_name = self._name + "_" + stat + "_" + axis_str
        notation = notations.add_operator(self._notation, stat, axis=axis_str)

        if axis == (0, 1):
            return StaticResults0D(new_name, dnew, notation=notation)
        else:
            if axis == 0:
                coords = self._x
                coords_name = "x"
            else:
                coords = self._y
                coords_name = "y"
            return StaticResults1D(
                new_name,
                dnew,
                coords,
                coords_name=coords_name,
                notation=notation,
            )

    @property
    def x(self) -> np.ndarray:
        """Get X coordinates.

        Returns
        -------
        numpy.ndarray
            Attribute :attr:`_x`.
        """
        return self._x

    @property
    def y(self) -> np.ndarray:
        """Get Y coordinates.

        Returns
        -------
        numpy.ndarray
            Attribute :attr:`_y`.
        """
        return self._y

    @property
    def z(self) -> np.ndarray:
        """Get elevations values.

        Returns
        -------
        numpy.ndarray
            Attribute :attr:`_z`.
        """
        return self._z


class Results:
    """Results of thin-layer model simulation

    This class is the parent class for all simulation results, whatever the
    kind of input data. Methods and functions for processing results are given
    here. Reading results from code specific outputs is done in inhereited
    classes.

    This class has global attributes used by all child classes and quick access
    attributes calculated and stored for easier access to the main results of a
    simulation. The quick attributes are only computed if needed and can be deleted
    to clean memory.

    Parameters
    ----------
    args and kwargs :
        Specific arguments for each models.

    Attributes
    ----------
        _code : str
            Name of the code that generated the result.
        _folder : str
            Path to find code files (like parameters).
        _folder_output :
            Path to find the results of the code.
        _zinit : numpy.ndarray
            Surface elevation of the simulation.
        _tim : list
            Lists of recorded time steps.
        _x : numpy.ndarray
            X-coordinates of the simulation.
        _y : numpy.ndarray
            Y-coordinates of the simulation.
        _dx : float
            Cell size along X-coordinates.
        _dy : float
            Cell size along Y-coordinates.
        _nx : float
            Number of cells along X-coordinates.
        _ny : float
            Number of cells along Y-coordinates.

        _h : tilupy.read.TemporalResults2D
            Quick access attributes for fluid height over time.
        _h_max : tilupy.read.TemporalResults0D
            Quick access attributes for max fluid hieght over time.
        _u : tilupy.read.TemporalResults2D
            Quick access attributes for norm of fluid velocity over time.
        _u_max : tilupy.read.TemporalResults0D
            Quick access attributes for max norm of fluid velocity over time.
        _costh : numpy.ndarray
            Quick access attributes for value of cos[theta] at any point on the surface.
    """

    def __init__(self, *args, **kwargs):
        self._h = None
        self._h_max = None
        self._u = None
        self._u_max = None
        self._costh = None

        self._code = None
        self._folder = None
        self._folder_output = None
        self._z = None
        self._zinit = None
        self._tim = None
        self._x = None
        self._y = None
        self._dx = None
        self._dy = None
        self._nx = None
        self._ny = None

    def compute_costh(self) -> np.ndarray:
        """Get cos(slope) of topography.

        Returns
        -------
        numpy.ndarray
            Value of cos[theta] at any point on the surface.
        """
        [Fx, Fy] = np.gradient(self._zinit, np.flip(self._y), self._x)
        costh = 1 / np.sqrt(1 + Fx**2 + Fy**2)
        return costh

    def center_of_mass(self, h_thresh: float = None) -> tilupy.read.TemporalResults0D:
        """Compute center of mass coordinates depending on time.

        Parameters
        ----------
        h_thresh : float, optional
            Value of threshold for the flow height, by default None.

        Returns
        -------
        tilupy.read.TemporalResults0D
            Values of center of mass coordinates.
        """
        # Make meshgrid
        X, Y = np.meshgrid(self._x, np.flip(self._y))
        Z = self.z.copy()

        # Ignore thicknesses below given threshold
        h2 = self.h.copy()
        if h_thresh is not None:
            h2[h2 < h_thresh] = 0

        # Get the coordinates of the center of mass at each cell.
        # It corresponds to the cell coordinates + half the thickness in the direction normal to topograpy
        # COmpute the unit vector perpendicular to the topography
        nx, ny, nz = tilupy.utils.normal_vector(self.x, self.y, self.z)
        X = X[:, :, np.newaxis] + 0.5 * h2 * nx[:, :, np.newaxis]
        Y = Y[:, :, np.newaxis] + 0.5 * h2 * ny[:, :, np.newaxis]
        Z = Z[:, :, np.newaxis] + 0.5 * h2 * nz[:, :, np.newaxis]

        # Weights for coordinates average (volume in cell / total volume)
        w = (
            h2 / self.costh[:, :, np.newaxis]
        )  # Multiplication by dx*dy is useless as it then simplies
        vol = np.nansum(w, axis=(0, 1))
        w = w / vol[np.newaxis, np.newaxis, :]
        # Compute center of mass coordinates
        nt = h2.shape[2]
        coord = np.zeros((3, nt))
        coord[0, :] = np.nansum(X * w, axis=(0, 1))
        coord[1, :] = np.nansum(Y * w, axis=(0, 1))
        coord[2, :] = np.nansum(Z * w, axis=(0, 1))

        # Make TemporalResults
        res = TemporalResults0D(
            "centermass", coord, self._tim, scalar_names=["X", "Y", "z"], notation=None
        )
        return res

    def volume(self, h_thresh: float = None) -> tilupy.read.TemporalResults0D:
        """Compute flow volume depending on time.

        Parameters
        ----------
        h_thresh : float, optional
            Value of threshold for the flow height, by default None.

        Returns
        -------
        tilupy.read.TemporalResults0D
            Values of flow volumes.
        """
        dx = self._x[1] - self._x[0]
        dy = self._y[1] - self._y[0]

        if self._h is None:
            self._h = self.get_output("h").d
        h2 = self._h.copy()
        if h_thresh is not None:
            h2[h2 < h_thresh] = 0
        if self._costh is None:
            self._costh = self.compute_costh()
        w = h2 / self._costh[:, :, np.newaxis] * dx * dy
        vol = np.nansum(w, axis=(0, 1))
        res = TemporalResults0D("volume", vol, self._tim, notation=None)
        return res

    def get_output(
        self, output_name: str, from_file: bool = True, **kwargs
    ) -> (
        tilupy.read.TemporalResults0D
        | tilupy.read.StaticResults2D
        | tilupy.read.TemporalResults2D
    ):
        """Get all the available outputs for a simulation :
            - Topographic outputs : "z", "zinit", "costh"
            - Temporal 2D outputs : "hvert", "h", "u", "ux", "uy", "hu", "hu2"
            - Other outputs : "centermass", "volume"

        It is possible to add operators to temporal 2D outputs :
            - "max", "mean", "std", "sum", "min", "final", "init", "int"

        And it is possible to add axis (only if using operators) :
            - "x", "y", "xy"

        Parameters
        ----------
        output_name : str
            Name of the wanted output, composed of the output name and potentially
            an operator and an axis: :data:`output_operator_axis`.
        from_file : bool, optional
            If True, find the output in a specific file. By default True.

        Returns
        -------
        tilupy.read.TemporalResults0D, tilupy.read.StaticResults2D or tilupy.read.TemporalResults2D
            Wanted output.
        """
        # Specific case of center of mass
        if output_name == "centermass":
            return self.center_of_mass(**kwargs)

        # Specific case of volume
        if output_name == "volume":
            return self.volume(**kwargs)

        strs = output_name.split("_")
        n_strs = len(strs)

        res = None

        # get topography
        if output_name in TOPO_DATA_2D:
            if output_name == "z":
                output_name = "zinit"
            output_name = "_" + output_name
            res = StaticResults2D(
                output_name,
                getattr(self, output_name),
                x=self._x,
                y=self._y,
                z=self._z,
                notation=None,
            )
            return res

        # If no operator is called, call directly extract_output
        if n_strs == 1:
            res = self._extract_output(output_name, **kwargs)
            return res

        # Otherwise, get name, operator and axis (optional)
        name = strs[0]
        operator = strs[1]
        axis = None
        if n_strs == 3:
            axis = strs[2]

        # If processed output is read directly from file, call the child method
        # read_from_file.
        if from_file:
            try:
                res = self._read_from_file(name, operator, axis=axis, **kwargs)
                if res is None:
                    raise UserWarning(
                        f"{output_name} not found with _read_from_file for {self._code}, use get_spatial_stat"
                    )
                elif isinstance(res, str):
                    raise UserWarning(res)
            except UserWarning as w:
                print(f"[WARNING] {w}")
                res = None
            # res is None in case of function failure

        # If no results could be read from file, output must be
        # processed by tilupy
        if res is None:
            # Get output from name
            res = self._extract_output(name, x=self._x, y=self._y, **kwargs)
            if axis is None:
                # If no axis is given, the operator operates over time by
                # default
                res = res.get_temporal_stat(operator)
            else:
                if axis == "t":
                    res = res.get_temporal_stat(operator)
                else:
                    res = res.get_spatial_stat(operator, axis=axis)

        return res

    def clear_quick_results(self) -> None:
        """Clear memory by erasing quick access attributes: :attr:`_h`, :attr:`_h_max`, :attr:`_u`, :attr:`_u_max`, :attr:`_costh`."""
        self._h = None
        self._h_max = None
        self._u = None
        self._u_max = None
        self._costh = None

    def get_profile(
        self, output: str, extraction_method: str = "axis", **extraction_params
    ) -> tuple[tilupy.read.TemporalResults1D | tilupy.read.StaticResults1D, np.ndarray]:
        """Extract a profile from a 2D data.

        Parameters
        ----------
        output : str
            Wanted data output.
        extraction_mode : str, optional
            Method to extract profiles:

                - "axis": Extracts a profile along an axis.
                - "coordinates": Extracts a profile along specified coordinates.
                - "shapefile": Extracts a profile along a shapefile (polylines).

            Be default "axis".
        extraction_params : dict, optional
            Different parameters to be entered depending on the extraction method chosen.
            See :meth:`tilupy.utils.get_profile`.

        Returns
        -------
        tuple[tilupy.read.TemporalResults1D | tilupy.read.StaticResults1D, np.ndarray]
            profile : tilupy.read.TemporalResults1D | tilupy.read.StaticResults1D
                Extracted profile.
            data : numpy.ndarray or float or tuple[nympy.ndarray, nympy.ndarray]
                Specific output depending on :data:`extraction_mode`:

                    - If :data:`extraction_mode == "axis"`: float
                        Position of the profile.
                    - If :data:`extraction_mode == "coordinates"`: tuple[numpy.ndarray]
                        X coordinates, Y coordinates and distance values.
                    - If :data:`extraction_mode == "shapefile"`: numpy.ndarray
                        Distance values.

        Raises
        ------
        ValueError
            If :data:`output` doesn't generate a 2D data.
        """
        data = self.get_output(output)

        if not isinstance(data, tilupy.read.TemporalResults2D) and not isinstance(
            data, tilupy.read.StaticResults2D
        ):
            raise ValueError("Can only extract profile from 2D data.")

        profile, data = utils.get_profile(data, extraction_method, **extraction_params)

        return profile, data

    def plot(
        self,
        output: str,
        from_file: bool = True,  # get_output
        h_thresh: float = None,  # get_output
        time_steps: float | list[float] = None,
        save: bool = False,
        folder_out: str = None,
        dpi: int = 150,
        fmt: str = "png",
        file_suffix: str = None,
        file_prefix: str = None,
        display_plot: bool = True,
        **plot_kwargs,
    ) -> matplotlib.axes._axes.Axes:
        """Plot output extracted from model's result.

        Parameters
        ----------
        output : str
            Wanted output to be plotted. Must be in :data:`DATA_NAMES`.
        from_file : bool, optional
            If True, find the output in a specific file. By default True.
        h_thresh : float, optional
            Threshold value to be taken into account when extracting output, by default None.
        time_steps : float or list[float], optional
            Time steps to show when plotting temporal data. If None shows every time
            steps recorded. By default None.
        save : bool, optional
            If True, save the plot as an image to the computer, by default False.
        folder_out : str, optional
            Path to the folder where to save the plot, by default None.
        dpi : int, optional
            Resolution for the saved plot, by default 150.
        fmt : str, optional
            Format of the saved plot, by default "png".
        file_suffix : str, optional
            Suffix to add to the file name when saving, by default None.
        file_prefix : str, optional
            Prefix to add to the file name when saving, by default None.
        display_plot : bool, optional
            If True, enables the display of the plot; otherwise, it disables the display to save memory.
            By default True.

        Returns
        -------
        matplotlib.axes._axes.Axes
            Wanted plot.
        """
        if not display_plot:
            backend = plt.get_backend()
            plt.close("all")
            plt.switch_backend("Agg")

        if output in ["z", "zinit", "z_init"]:
            topo_kwargs = dict()
            if "topo_kwargs" in plot_kwargs:
                topo_kwargs = plot_kwargs["topo_kwargs"]
            axe = plt_fn.plot_topo(self._zinit, self._x, self._y, **topo_kwargs)
            return axe

        data = self.get_output(output, from_file=from_file, h_thresh=h_thresh)

        add_time_on_plot = False
        if (
            isinstance(time_steps, float) or isinstance(time_steps, int)
        ) and isinstance(data, tilupy.read.TemporalResults):
            t_index = np.argmin(np.abs(self._tim - time_steps))
            add_time_on_plot = self._tim[t_index]

        if time_steps is not None and isinstance(data, tilupy.read.TemporalResults):
            data = data.extract_from_time_step(time_steps)

        if save:
            if folder_out is None:
                assert self._folder_output is not None, (
                    "folder_output attribute must be set"
                )
                folder_out = os.path.join(self._folder_output, "plots")
            os.makedirs(folder_out, exist_ok=True)

        # TODO
        # Edit Temporal/Static.plot() pour que la sauvegarde soit directement intégrer dans les méthodes plots
        # Dans les plots, modifier les fonctions pour appeler des méthodes de pytopomap selon le plot voulu
        # (shotgater, profil, surface, etc...) et créer une fonction globale qui appelle chaque sous méthode
        # pour créer les graphes et gérer la sauvegarde
        if folder_out is not None and isinstance(data, TemporalResults2D):
            # If data is TemporalResults2D then saving is managed directly
            # by the associated plot method
            plot_kwargs["folder_out"] = folder_out
            plot_kwargs["dpi"] = dpi
            plot_kwargs["fmt"] = fmt
            # kwargs["file_suffix"] = file_prefix
            # kwargs["file_prefix"] = file_prefix

        axe = data.plot(**plot_kwargs)

        if add_time_on_plot:
            axe.set_title(f"t={add_time_on_plot}s", loc="left")

        if folder_out is not None and not isinstance(data, TemporalResults2D):
            file_name = output
            if file_suffix is not None:
                file_name = file_name + "_" + file_suffix
            if file_prefix is not None:
                file_name = file_prefix + "_" + file_name
            file_out = os.path.join(folder_out, file_name + "." + fmt)
            # axe.figure.tight_layout(pad=0.1)
            axe.figure.savefig(file_out, dpi=dpi, bbox_inches="tight", pad_inches=0.05)

        if not display_plot:
            plt.close("all")
            plt.switch_backend(backend)

        return axe

    def plot_profile(
        self,
        output: str,
        from_file: bool = True,
        extraction_method: str = "axis",
        extraction_params: dict = None,
        time_steps: float | list[float] = None,
        save: bool = False,
        folder_out: str = None,
        display_plot: bool = True,
        **plot_kwargs,
    ) -> matplotlib.axes._axes.Axes:
        """Plot a 1D output extracted from a 2D output.

        Parameters
        ----------
        output : str
            Wanted 2D output to extract the profile from. Must be in :data:`STATIC_DATA_2D`
            or in :data:`TEMPORAL_DATA_2D`.
        from_file : bool, optional
            If True, find the output in a specific file. By default True.
        extraction_mode : str, optional
            Method to extract profiles:

                - "axis": Extracts a profile along an axis.
                - "coordinates": Extracts a profile along specified coordinates.
                - "shapefile": Extracts a profile along a shapefile (polylines).
            Be default "axis".
        extraction_params : dict, optional
            Different parameters to be entered depending on the extraction method chosen.
            See :meth:`tilupy.utils.get_profile`.
        time_steps : float or list[float], optional
            Time steps to show when plotting temporal data. If None shows every time
            steps recorded. By default None.
        save : bool, optional
            If True, save the plot as an image to the computer, by default False.
        folder_out : str, optional
            Path to the folder where to save the plot, by default None.
        display_plot : bool, optional
            If True, enables the display of the plot; otherwise, it disables the display to save memory.
            By default True.

        Returns
        -------
        matplotlib.axes._axes.Axes
            Wanted plot.

        Raises
        ------
        ValueError
            If the :data:`output` is not a 2D output.
        """
        if not display_plot:
            backend = plt.get_backend()
            plt.close("all")
            plt.switch_backend("Agg")

        data = self.get_output(output, from_file=from_file)

        if not isinstance(data, tilupy.read.TemporalResults2D) and not isinstance(
            data, tilupy.read.StaticResults2D
        ):
            raise ValueError("Can only extract profile from 2D data.")

        extraction_params = {} if extraction_params is None else extraction_params

        profile, _ = utils.get_profile(data, extraction_method, **extraction_params)
        closest_value = False

        if (
            isinstance(time_steps, float) or isinstance(time_steps, int)
        ) and isinstance(data, tilupy.read.TemporalResults):
            t_index = np.argmin(np.abs(self._tim - time_steps))
            closest_value = self._tim[t_index]

        if time_steps is not None and isinstance(data, tilupy.read.TemporalResults):
            profile = profile.extract_from_time_step(time_steps)

        axe = profile.plot(**plot_kwargs)

        if closest_value:
            axe.set_title(f"t={closest_value}s", loc="left")

        if save:
            if folder_out is None:
                assert self._folder_output is not None, (
                    "folder_output attribute must be set"
                )
                folder_out = os.path.join(self._folder_output, "plots")
            os.makedirs(folder_out, exist_ok=True)

        # TODO
        # Same as plot() -> add save mode in plot functions in Temporal/StaticResults
        """
        if folder_out is not None and isinstance(data, TemporalResults2D):
            # If data is TemporalResults2D then saving is managed directly
            # by the associated plot method
            kwargs["folder_out"] = folder_out
            kwargs["dpi"] = dpi
            kwargs["fmt"] = fmt
            # kwargs["file_suffix"] = file_prefix
            # kwargs["file_prefix"] = file_prefix

        

        if folder_out is not None and not isinstance(data, TemporalResults2D):
            file_name = output_name
            if file_suffix is not None:
                file_name = file_name + "_" + file_suffix
            if file_prefix is not None:
                file_name = file_prefix + "_" + file_name
            file_out = os.path.join(folder_out, file_name + "." + fmt)
            # axe.figure.tight_layout(pad=0.1)
            axe.figure.savefig(file_out, dpi=dpi, bbox_inches="tight", pad_inches=0.05)
        """

        if not display_plot:
            plt.close("all")
            plt.switch_backend(backend)

        return axe

    def save(
        self,
        output_name: str,
        folder_out: str = None,
        file_name: str = None,
        fmt: str = "txt",
        from_file: bool = True,
        **kwargs,
    ) -> None:
        """Save simulation outputs (processed results or topographic data) to disk.

        Depending on the requested output_name, the method either:

            - Retrieves a result via :meth:`get_output` and calls its own :meth:`save` method,
            - Or, for static topography data, writes it directly to a raster file.

        Parameters
        ----------
        output_name : str
            Name of the variable or processed result to save
            (e.g., "h", "u_mean_t", "centermass", or topographic data like "zinit").
        folder_out : str, optional
            Destination folder for saving files. If None, defaults to
            :data:`_folder_output/processed`. By default None.
        file_name : str, optional
            Base name of the output file (without extension). If None, uses
            :data:`output_name`. By default None.
        fmt : str, optional
            Output file format (e.g., "txt", "npy", "asc"), by default "txt".
        from_file : bool, optional
            If True, attempt to read precomputed results from file before computing,
            by default True.
        **kwargs : dict
            Extra arguments passed to the underlying save function. For raster data,
            forwarded to :func:`tilupy.raster.write_raster`.

        Raises
        ------
        AssertionError
            If neither :data:`folder_out` nor :attr:`_folder_output` is defined.
        """

        if folder_out is None:
            assert self._folder_output is not None, (
                "folder_output attribute must be set"
            )
            folder_out = os.path.join(self._folder_output, "processed")
            os.makedirs(folder_out, exist_ok=True)

        if output_name in DATA_NAMES:
            data = self.get_output(output_name, from_file=from_file)
            if data.d.ndim > 1:
                if "x" not in kwargs:
                    kwargs["x"] = self.x
                if "y" not in kwargs:
                    kwargs["y"] = self.y

            data.save(folder=folder_out, file_name=file_name, fmt=fmt, **kwargs)

        elif output_name in TOPO_DATA_2D:
            if file_name is None:
                file_name = output_name
            name = "_" + output_name
            file_out = os.path.join(folder_out, file_name)
            tilupy.raster.write_raster(
                self._x, self._y, getattr(self, name), file_out, fmt=fmt, **kwargs
            )

    @abstractmethod
    def _extract_output(self):
        """Abstract method to extract output of simulation result files."""
        pass

    @abstractmethod
    def _read_from_file(self):
        """Abstract method for reading output from specific files."""
        pass

    @property
    def zinit(self):
        """Get initial topography.

        Returns
        -------
        numpy.ndarray
            Attribute :attr:`_zinit`
        """
        return self._zinit

    @property
    def z(self):
        """Get initial topography, alias for zinit.

        Returns
        -------
        numpy.ndarray
            Attribute :attr:`_zinit`
        """
        return self._zinit

    @property
    def x(self):
        """Get X-coordinates.

        Returns
        -------
        numpy.ndarray
            Attribute :attr:`_x`
        """
        return self._x

    @property
    def y(self):
        """Get Y-coordinates.

        Returns
        -------
        numpy.ndarray
            Attribute :attr:`_y`
        """
        return self._y

    @property
    def dx(self):
        """Get cell size along X.

        Returns
        -------
        numpy.ndarray
            Attribute :attr:`_dx`
        """
        return self._dx

    @property
    def dy(self):
        """Get cell size along Y.

        Returns
        -------
        numpy.ndarray
            Attribute :attr:`_dy`
        """
        return self._dy

    @property
    def nx(self):
        """Get number of cells along X.

        Returns
        -------
        numpy.ndarray
            Attribute :attr:`_nx`
        """
        return self._nx

    @property
    def ny(self):
        """Get number of cells along Y.

        Returns
        -------
        numpy.ndarray
            Attribute :attr:`_ny`
        """
        return self._ny

    @property
    def tim(self):
        """Get recorded time steps.

        Returns
        -------
        numpy.ndarray
            Attribute :attr:`_tim`
        """
        return self._tim

    @property
    def h(self):
        """Get flow thickness. Compute it if not stored.

        Returns
        -------
        tilupy.read.TemporalResults2D
            Attribute :attr:`_h`
        """
        if self._h is None:
            self._h = self.get_output("h").d
        return self._h

    @property
    def h_max(self):
        """Get maximum flow thickness. Compute it if not stored.

        Returns
        -------
        tilupy.read.TemporalResults0D
            Attribute :attr:`_h_max`
        """
        if self._h_max is None:
            self._h_max = self.get_output("h_max").d
        return self._h_max

    @property
    def u(self):
        """Get flow velocity. Compute it if not stored.

        Returns
        -------
        tilupy.read.TemporalResults2D
            Attribute :attr:`_u`
        """
        if self._u is None:
            self._u = self.get_output("u").d
        return self._u

    @property
    def u_max(self):
        """Get maximum flow velocity. Compute it if not stored.

        Returns
        -------
        tilupy.read.TemporalResults0D
            Attribute :attr:`_u_max`
        """
        if self._u_max is None:
            self._u_max = self.get_output("u_max").d
        return self._u_max

    @property
    def costh(self):
        """Get cos(slope) of topography. Compute it if not stored.

        Returns
        -------
        numpy.ndarray
            Attribute :attr:`_costh`
        """
        if self._costh is None:
            self._costh = self.compute_costh()
        return self._costh


def get_results(code, **kwargs) -> tilupy.read.Results:
    """Get simulation results for a given numerical model.

    Dynamically imports the corresponding reader module from
    `tilupy.models.<code>.read` and instantiates its :class:`tilupy.read.Results` class.

    Parameters
    ----------
    code : str
        Short name of the simulation model: must be in :data:`ALLOWED_MODELS`.
    **kwargs : dict
        Additional keyword arguments passed to the :class:`tilupy.read.Results` constructor
        of the imported module.

    Returns
    -------
    tilupy.read.Results
        Instance of the :class:`tilupy.read.Results` class containing the simulation outputs.

    Raises
    ------
    ModuleNotFoundError
        If the module `tilupy.models.<code>.read` cannot be imported.
    AttributeError
        If the module does not define a :class:`tilupy.read.Results` class.
    """
    module = importlib.import_module("tilupy.models." + code + ".read")
    return module.Results(**kwargs)


def use_thickness_threshold(
    simu: tilupy.read.Results, array: np.ndarray, h_thresh: float
) -> np.ndarray:
    """Apply a flow thickness threshold to mask simulation results.

    Values of :data:`array` are set to zero wherever the flow thickness
    is below the given threshold.

    Parameters
    ----------
    simu : tilupy.read.Results
        Simulation result object providing access to thickness data :data:`h`.
    array : numpy.ndarray
        Array of values to be masked (must be consistent in shape with thickness).
    h_thresh : float
        Thickness threshold. Cells with thickness < :data:`h_thresh` are set to zero.

    Returns
    -------
    numpy.ndarray
        Thresholded array, with values set to zero where flow thickness is too low.
    """
    thickness = simu.get_output("h")
    array[thickness.d < h_thresh] = 0
    return array



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import numpy as np

import tilupy.make_topo
import tilupy.make_mass
import tilupy.raster


def create_topo_constant_slope(
    folder_out: str,
    xmax: int = 30,
    ymax: int = 25,
    cell_size: float = 0.5,
    theta: int = 5,
    mass_type: str = "r",
    r_center: tuple = (7.5, 12.5),
    r_radius: tuple = (3.75, 3.75),
    s_vertex: list = [0, 5, 0, 25],
    h_max: float = 3.75,
    description: str = "No informations.",
) -> list[str, np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """
    Generates a synthetic topography and initial mass, saves them as ASCII files,
    and stores all configuration parameters in a dedicated folder.

    Parameters
    ----------
    folder_out : str
        Output folder.
    xmax : int
        Domain length in the X direction (in meters).
    ymax : int
        Domain length in the Y direction (in meters).
    cell_size : float
        Spatial resolution of the grid (in meters).
    theta : int
        Slope angle of the topography (in degrees).
    mass_type : str
        Shape of the initial mass: 'r' for ellipse, 's' for rectangle.
    r_center : tuple
        Center (x, y) of the elliptical initial mass.
    r_radius : tuple
        Radii (rx, ry) of the elliptical initial mass.
    s_vertex : list
        Rectangle boundaries [xmin, xmax, ymin, ymax] for the rectangular mass.
    h_max : float
        Maximum height of the initial mass.
    description : str
        Text description of the configuration.

    Returns
    -------
    list[str, numpy.ndarray, numpy.ndarray, numpy.ndarray, numpy.ndarray]
        folder_path : str
            Absolute path to the output directory.
        x, y : numpy.ndarray, numpy.ndarray
            Meshgrid coordinates.
        z : numpy.ndarray
            Topography data.
        m : numpy.ndarray
            Initial mass distribution.
    """
    # Create mesh
    x = np.linspace(0, xmax, int(np.round(xmax / cell_size)) + 1)
    y = np.linspace(0, ymax, int(np.round(ymax / cell_size)) + 1)
    xmesh, ymesh = np.meshgrid(x, y)

    # Topography slope
    slope = np.tan(np.deg2rad(theta))

    # Topography array
    z = -slope * (xmesh - xmax)

    # Initial mass
    if mass_type != "r" and mass_type != "s":
        mass_type = "r"

    p = dict(
        hmax=h_max,
        mass_type=mass_type,
        r_center=r_center if mass_type == "r" else None,
        r_radius=r_radius if mass_type == "r" else None,
        s_vertex=s_vertex if mass_type == "s" else None,
    )

    if p["mass_type"] == "r":
        m = (
            1
            - (xmesh - p["r_center"][0]) ** 2 / p["r_radius"][0] ** 2
            - (ymesh - p["r_center"][1]) ** 2 / p["r_radius"][1] ** 2
        )
        m = np.maximum(m * p["hmax"], 0)

    elif p["mass_type"] == "s":
        xmin_r, xmax_r, ymin_r, ymax_r = p["s_vertex"]
        in_x = np.logical_and(xmesh >= xmin_r, xmesh <= xmax_r)
        in_y = np.logical_and(ymesh >= ymin_r, ymesh <= ymax_r)
        mask = np.logical_and(in_x, in_y)
        m = np.zeros_like(z)
        m[mask] = p["hmax"]

    nbr_cell = np.count_nonzero(m > 0)

    p["nbr_cell"] = nbr_cell

    folder_name = (
        f"x{xmax}_y{ymax}_"
        + "dx{:04.2f}_".format(cell_size).replace(".", "p")
        + "theta{:02.0f}_".format(theta)
        + mass_type
        + f"{nbr_cell}"
    )
    folder_path = os.path.join(folder_out, folder_name)

    # Create folder
    if not os.path.isdir(folder_path):
        os.mkdir(folder_path)
        print(f"Create folder: {folder_name}")
    else:
        print(f"Existing topography: {folder_name}")

    # Create parameters file
    with open(os.path.join(folder_path, "parameters.txt"), "w") as file:
        file.write(
            f"xmax {xmax}\nymax {ymax}\ncell_size {cell_size}\ntheta {theta}\n\n"
        )

        for param, value in p.items():
            file.write(f"{param} {value}\n")

        file.write("\n")
        file.write(description)

    # Save topography
    file_topo_out = os.path.join(folder_path, "topography.asc")
    file_mass_out = os.path.join(folder_path, "init_mass.asc")

    tilupy.raster.write_ascii(x, y, z, file_topo_out)
    tilupy.raster.write_ascii(x, y, m, file_mass_out)

    return folder_path, x, y, z, m


def gray99_topo_mass(
    dx: float = 0.1, dy: float = 0.1, res_type: str = "true_normal"
) -> list[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """Create Gray99 topographic and mass test.

    Create an initial spherical calotte above the topography, resulting as a mass with a
    height of 0.22 m and a radius of 0.32 m (more precisely it is the length in the downslope
    direction), following the indications in Gray et al 99 (p. 1859).
    The correspondig radius of the sphere, and the offset from the topography in the topography
    normal direction (norm_offset) are deduced from these parameters.

    See also Gig 3 in Wieland, Gray and Hutter (1999)

    Parameters
    ----------
    dx : float, optional
        Cell size of the x axis, by default 0.1
    dy : float, optional
        Cell size of the y axis, by default 0.1
    res_type : str, optional
        Type of thickness output:

            - 'true_normal': Real thickness in the direction normal to the topography.
            - 'vertical': Thickness in the vertical direction.
            - 'projected_normal': Thickness normal to the topography is computed from the vertical
              thickness projected on the axe normal to the topography.

        The default is 'true_normal'.

    Returns
    -------
    list[numpy.ndarray, numpy.ndarray, numpy.ndarray, numpy.ndarray]
        X : numpy.ndarray
            Mesh of X coordinates in the cartesian frame (nx*ny).
        Y : numpy.ndarray
            Mesh of Y coordinates in the cartesian frame (nx*ny).
        Z : numpy.ndarray
            Mesh of Z coordinates in the cartesian frame (nx*ny).
        M : numpy.ndarray
            Array of mass height, in the direction normal to topography.
    """
    # Initiate topography
    X, Y, Z = tilupy.make_topo.gray99(dx=dx, dy=dy)

    x0 = 0.06 * np.cos(np.deg2rad(40))
    hmass = 0.22
    wmass = 0.32
    radius = (wmass**2 + hmass**2) / (2 * hmass)
    norm_offset = (wmass**2 - hmass**2) / (2 * hmass)
    # Z = -np.tile(X, [len(Y), 1])*np.tan(np.deg2rad(20))

    M = tilupy.make_mass.calotte(
        X, Y, Z, x0, 0, radius, norm_offset=norm_offset, res_type=res_type
    )

    return X, Y, Z, M


"""
if __name__ == "__main__":
    x, y, z, m = gray99_topo_mass(dx=0.01, dy=0.01)
    axe = pytopomap.plot.plot_data_on_topo(x, y, z, m, topo_kwargs=dict(level_min=0.1))
    plt.show()
"""



# -*- coding: utf-8 -*-

from __future__ import annotations

import numpy as np
import matplotlib.pyplot as plt

import shapely.geometry as geom
import shapely.ops

import tilupy.read


def normal_vector(
    x: np.ndarray, y: np.ndarray, z: np.ndarray
) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Compute unit vector normal to the topography

    Return x, y, and z coordinated of the unit vector normal to the topography oriented upwards

    Parameters
    ----------
    x : numpy.ndarray
        Array of length N with x-coordinates
    y : numpy.ndarray
        Array of length M with y-coordinates
    z : numpy.ndarray
        MxN array of topography. z[0,0] has coordinates x[0], y[-1],
        z[-1, -1] has coordinates x[-1], y[0]

    Returns
    -------
    nx : numpy.ndarray
        MxN array with x coordinate of the unit upward normal vector
    ny : numpy.ndarray
        MxN array with y coordinate of the unit upward normal vector
    nz : numpy.ndarray
        MxN array with z coordinate of the unit upward normal vector
    """

    [Fy, Fx] = np.gradient(z, y, x)
    norm = np.sqrt(1 + Fx**2 + Fy**2)
    nz = 1 / norm
    nx = -Fx / norm
    ny = -Fy / norm
    return nx, ny, nz


def CSI(pred: np.ndarray, obs: np.ndarray) -> float:
    """Compute the Critical Success Index (CSI).

    Measure the fraction of observed and/or predicted events that were correctly predicted.

    Parameters
    ----------
    pred : numpy.ndarray
        Array of predicted binary values (e.g., 1 for event predicted, 0 otherwise).
        Shape and type must match :data:`obs`.
    obs : numpy.ndarray
        Array of observed binary values (e.g., 1 for event observed, 0 otherwise).
        Shape and type must match :data:`pred`.

    Returns
    -------
    float
        The Critical Success Index (CSI), a score between 0 and 1.
        A score of 1 indicates perfect prediction, while 0 indicates no skill.
    """
    ipred = pred > 0
    iobs = obs > 0

    TP = np.sum(ipred * iobs)
    FP = np.sum(ipred * ~iobs)
    FN = np.sum(~ipred * iobs)

    return TP / (TP + FP + FN)


def diff_runout(
    x_contour: np.ndarray,
    y_contour: np.ndarray,
    point_ref: tuple[float, float],
    section: np.ndarray | shapely.geometry.LineString = None,
    orientation: str = "W-E",
) -> float:
    """Compute runout distance difference between a reference point and its projection on a contour,
    optionally along a specified section line.

    This function calculates the distance from a reference point to a contour (e.g., a polygon boundary),
    or the distance along a section line (e.g., a transect) between the reference point and its intersection
    with the contour. The section line can be oriented in four cardinal directions.

    Parameters
    ----------
    x_contour : numpy.ndarray
        Array of x-coordinates defining the contour.
    y_contour : numpy.ndarray
        Array of y-coordinates defining the contour. Must be the same length as :data:`x_contour`.
    point_ref : tuple[float, float]
        Reference point coordinates (x, y) from which the distance is calculated.
    section : numpy.ndarray or shapely.geometry.LineString, optional
        Coordinates of the section line as an array of shape (N, 2) or a Shapely LineString.
        If None, the function returns the Euclidean distance from :data:`point_ref` to the contour.
        By default None.
    orientation : str, optional
        Orientation of the section line. Must be one of:

            - "W-E" (West-East, default)
            - "E-W" (East-West)
            - "S-N" (South-North)
            - "N-S" (North-South)

        This determines how the intersection point is selected if multiple intersections exist.
        By default "W-E".

    Returns
    -------
    float
        If `section` is None: Euclidean distance from :data:`point_ref` to the contour.
        If `section` is provided: Distance along the section line between :data:`point_ref` and the contour intersection.
        The distance is signed, depending on the projection direction.
    """
    npts = len(x_contour)
    contour = geom.LineString([(x_contour[i], y_contour[i]) for i in range(npts)])
    point = geom.Point(point_ref)
    if section is None:
        return point.distance(contour)
    elif isinstance(section, np.ndarray):
        section = geom.LineString(section)

    assert isinstance(section, geom.LineString)
    section = revert_line(section, orientation)
    intersections = section.intersection(contour)
    if isinstance(intersections, geom.MultiPoint):
        intersections = geom.LineString(intersections.geoms)
    intersections = np.array(intersections.coords)
    if orientation == "W-E":
        i = np.argmax(intersections[:, 0])
    if orientation == "E-W":
        i = np.argmin(intersections[:, 0])
    if orientation == "S-N":
        i = np.argmax(intersections[:, 1])
    if orientation == "N-S":
        i = np.argmin(intersections[:, 1])
    intersection = geom.Point(intersections[i, :])

    #######
    # plt.figure()
    # cont = np.array(contour.coords)
    # sec = np.array(section.coords)
    # inter = np.array(intersection.coords)
    # pt = np.array(point.coords)
    # plt.plot(cont[:,0], cont[:,1],
    #           sec[:,0], sec[:,1],
    #           pt[:,0], pt[:,1], 'o',
    #           inter[:,0], inter[:,1],'x')
    #######

    return section.project(intersection) - section.project(point)


def revert_line(
    line: shapely.geometry.LineString, orientation: str = "W-E"
) -> shapely.geometry.LineString:
    """Revert a line geometry if its orientation does not match the specified direction.

    This function checks the orientation of the input line (based on the coordinates of its first and last points)
    and reverses it if necessary to ensure it matches the specified cardinal direction.
    The line is reversed in-place if the initial orientation is opposite to the specified one.

    Parameters
    ----------
    line : shapely.geometry.LineString
        Input line geometry to be checked and potentially reverted.
    orientation : str, optional
        Desired cardinal direction for the line. Must be one of:

            - "W-E" (West-East, default): The line should start at the west end (smaller x-coordinate).
            - "E-W" (East-West): The line should start at the east end (larger x-coordinate).
            - "S-N" (South-North): The line should start at the south end (smaller y-coordinate).
            - "N-S" (North-South): The line should start at the north end (larger y-coordinate).
        By default "W-E".

    Returns
    -------
    shapely.geometry.LineString
        The input line, potentially reverted to match the specified orientation.
        If the line already matches the orientation, it is returned unchanged.
    """
    pt_init = line.coords[0]
    pt_end = line.coords[-1]
    if orientation == "W-E":
        if pt_init[0] > pt_end[0]:
            line = shapely.ops.substring(line, 1, 0, normalized=True)
    elif orientation == "E-W":
        if pt_init[0] < pt_end[0]:
            line = shapely.ops.substring(line, 1, 0, normalized=True)
    if orientation == "S-N":
        if pt_init[1] > pt_end[1]:
            line = shapely.ops.substring(line, 1, 0, normalized=True)
    elif orientation == "N-S":
        if pt_init[1] < pt_end[1]:
            line = shapely.ops.substring(line, 1, 0, normalized=True)

    return line


def get_contour(
    x: np.ndarray,
    y: np.ndarray,
    z: np.ndarray,
    zlevels: list[float],
    indstep: int = 1,
    maxdist: float = 30,
    closed_contour: bool = True,
) -> tuple[dict[float, np.ndarray], dict[float, np.ndarray]]:
    """Extract contour lines from a 2D grid of values at specified levels.

    This function computes contour lines for a given 2D field :data:`z` defined on the grid :data:`(x, y)`,
    at the specified levels :data:`zlevels`. It can optionally ensure that contours are closed by padding
    the input arrays, and filter out contours that are not closed within a maximum distance threshold.

    Parameters
    ----------
    x : numpy.ndarray
        1D array of x-coordinates defining the grid.
    y : numpy.ndarray
        1D array of y-coordinates defining the grid.
    z : numpy.ndarray
        2D array of values defined on the grid :data:`(x, y)`.
    zlevels : list[float]
        List of contour levels at which to extract the contours.
    indstep : int, optional
        Step used to subsample the contour points, by default 1 (no subsampling).
    maxdist : float, optional
        Maximum allowed distance between the first and last points of a contour line.
        If the distance exceeds this value and `closed_contour` is False, the contour is discarded,
        by default 30.
    closed_contour : bool, optional
        If True, the input arrays are padded to ensure closed contours, by default True.

    Returns
    -------
    tuple[dict[float, np.ndarray], dict[float, np.ndarray]]
        xcontour : dict
            Maps each contour level to an array of x-coordinates of the contour line.
        ycontour : dict
            Maps each contour level to an array of y-coordinates of the contour line.
        If a contour is discarded due to `maxdist`, the corresponding value is `None`.
    """
    # Add sup_value at the border of the array, to make sure contour
    # lines are closed
    if closed_contour:
        z2 = z.copy()
        ni = z2.shape[0]
        nj = z2.shape[1]
        z2 = np.vstack([np.zeros((1, nj)), z2, np.zeros((1, nj))])
        z2 = np.hstack([np.zeros((ni + 2, 1)), z2, np.zeros((ni + 2, 1))])
        dxi = x[1] - x[0]
        dxf = x[-1] - x[-2]
        dyi = y[1] - y[0]
        dyf = y[-1] - y[-2]
        x2 = np.insert(np.append(x, x[-1] + dxf), 0, x[0] - dxi)
        y2 = np.insert(np.append(y, y[-1] + dyf), 0, y[0] - dyi)
    else:
        x2, y2, z2 = x, y, z

    backend = plt.get_backend()
    plt.switch_backend("Agg")
    fig = plt.figure()
    ax = plt.gca()
    cs = ax.contour(x2, y2, np.flip(z2, 0), zlevels)
    nn1 = 1
    v1 = np.zeros((1, 2))
    xcontour = {}
    ycontour = {}
    for indlevel in range(len(zlevels)):
        levels = cs.allsegs[indlevel]
        for p in levels:
            if p.shape[0] > nn1:
                v1 = p
                nn1 = p.shape[0]
        xc = [v1[::indstep, 0]]
        yc = [v1[::indstep, 1]]
        if maxdist is not None and not closed_contour:
            ddx = np.abs(v1[0, 0] - v1[-1, 0])
            ddy = np.abs(v1[0, 1] - v1[-1, 1])
            dd = np.sqrt(ddx**2 + ddy**2)
            if dd > maxdist:
                xc[0] = None
                yc[0] = None
        xcontour[zlevels[indlevel]] = xc[0]
        ycontour[zlevels[indlevel]] = yc[0]
    plt.close(fig)
    plt.switch_backend(backend)
    return xcontour, ycontour


def get_profile(
    data: tilupy.read.TemporalResults2D | tilupy.read.StaticResults2D,
    extraction_mode: str = "axis",
    data_threshold: float = 1e-3,
    **extraction_params,
) -> tuple[
    tilupy.read.TemporalResults1D | tilupy.read.StaticResults1D,
    float | tuple[np.ndarray] | np.ndarray,
]:
    """Extract profile with different modes and options.

    Parameters
    ----------
    data : tilupy.read.TemporalResults2D or tilupy.read.StaticResults2D
        Data to extract the profile from.
    extraction_mode : str, optional
        Method to extract profiles:

            - "axis": Extracts a profile along an axis.
            - "coordinates": Extracts a profile along specified coordinates.
            - "shapefile": Extracts a profile along a shapefile (polylines).

        Be default "axis".

    data_threshold : float, optional
        Minimum value to consider as part of the profile, by default 1e-3.

    extraction_params : dict, optional
        Different parameters to be entered depending on the extraction method chosen:

            - If :data:`extraction_mode == "axis"`:

                - :data:`axis`: str, optional
                    Axis where to extract the profile ['X', 'Y'], by default 'Y'.
                - :data:`profile_position`: float, optional
                    Position where to extract the profile. If None choose the median.
                    By default None.
                Must be read: profile in :data:`axis = profile_position m`.

            - If :data:`extraction_mode == "coordinates"`:

                - :data:`xcoord`: numpy.ndarray, optional
                    X coordinates of the profile, by default :attr:`_x`.
                - :data:`ycoord`: numpy.ndarray, optional
                    Y coordinates of the profile, by default :data:`[0., 0., ...]`.

            - If :data:`extraction_mode == "shapefile"`:

                - :data:`path`: str
                    Path to the shapefile.
                - :data:`x_origin`: float, optional
                    Value of the X coordinate of the origin (top-left corner) in the shapefile's coordinate system, by default 0.0 (EPSG:2154).
                - :data:`y_origin`: float, optional
                    Value of the y coordinate of the origin (top-left corner) in the shapefile's coordinate system, by default :data:`_y[-1]` (EPSG:2154).
                - :data:`x_pixsize`: float, optional
                    Size of a pixel along the X coordinate in the shapefile's coordinate system, by default :data:`_x[1] - _x[0]` (EPSG:2154).
                - :data:`y_pixsize`: float, optional
                    Size of a pixel along the Y coordinate in the shapefile's coordinate system, by default :data:`_y[1] - _y[0]` (EPSG:2154).
                - :data:`step`: float, optional
                    Spatial step between profile points, by default 10.0.

        By default None

    Returns
    -------
    tuple[tilupy.read.TemporalResults1D or tilupy.read.StaticResults1D, float or tuple[np.ndarray] or np.ndarray]
        tilupy.read.TemporalResults1D or tilupy.read.StaticResults1D
            Extracted profiles.
        float or tuple[np.ndarray] or numpy.ndarray
            Specific output depending on :data:`extraction_mode`:

                - If :data:`extraction_mode == "axis"`: float
                    Position of the profile.
                - If :data:`extraction_mode == "coordinates"`: tuple[numpy.ndarray]
                    X coordinates, Y coordinates and distance values.
                - If :data:`extraction_mode == "shapefile"`: numpy.ndarray
                    Distance values.

    Raises
    ------
    ValueError
        If :data:`extraction_mode == "axis"` and if invalid :data:`axis`.
    ValueError
        If :data:`extraction_mode == "axis"` and if invalid format for :data:`profile_position`.
    ValueError
        If :data:`extraction_mode == "axis"` and if no value position found in axis.
    ValueError
        If :data:`extraction_mode == "coordinates"` and if invalid format for :data:`xcoord` or :data:`ycoord`.
    ValueError
        If :data:`extraction_mode == "coordinates"` and if invalid dimension for :data:`xcoord` or :data:`ycoord`.
    ValueError
        If :data:`extraction_mode == "coordinates"` and if :data:`xcoord` and :data:`ycoord` doesn't have same size.
    ValueError
        If :data:`extraction_mode == "shapefile"` and if no :data`path` is given.
    ValueError
        If :data:`extraction_mode == "shapefile"` and if invalid format for :data:`x_origin`, :data:`y_origin`, :data:`x_pixsize`, :data:`y_pixsize` or :data:`step`.
    TypeError
        If :data:`extraction_mode == "shapefile"` and if invalid geometry for the shapefile.
    ValueError
        If :data:`extraction_mode == "shapefile"` and if no linestring found in the shapefile.
    ValueError
        If invalid :data:`extraction_mode`.
    """
    if not isinstance(data, tilupy.read.TemporalResults2D) and not isinstance(
        data, tilupy.read.StaticResults2D
    ):
        raise ValueError("Can only extract profile from 2D data.")

    y_coord, x_coord = data.y, data.x
    (
        y_size,
        x_size,
    ) = len(y_coord), len(x_coord)
    data_field = data.d.copy()

    # Apply mask on data
    data_field[data_field <= data_threshold] = 0

    extraction_params = {} if extraction_params is None else extraction_params

    if extraction_mode == "axis":
        # Create specific params if not given
        if "axis" not in extraction_params:
            extraction_params["axis"] = "Y"
        if "profile_position" not in extraction_params:
            extraction_params["profile_position"] = None

        # Check errors
        if extraction_params["axis"] not in ["x", "X", "y", "Y"]:
            raise ValueError("Invalid axis: 'X' or 'Y'.")
        extraction_params["axis"] = extraction_params["axis"].upper()

        # Depending on "profile_position" type, choose median or position value
        if extraction_params["profile_position"] is None:
            if extraction_params["axis"] == "X":
                extraction_params["profile_index"] = x_size // 2
                closest_value = x_coord[extraction_params["profile_index"]]
            else:
                extraction_params["profile_index"] = y_size // 2
                closest_value = y_coord[extraction_params["profile_index"]]

        elif isinstance(extraction_params["profile_position"], float) or isinstance(
            extraction_params["profile_position"], int
        ):
            coord_val = extraction_params["profile_position"]
            x_index, y_index = None, None

            if extraction_params["axis"] == "X":
                if not isinstance(x_coord, np.ndarray):
                    x_coord = np.array(x_coord)

                x_index = np.argmin(np.abs(x_coord - coord_val))
                closest_value = x_coord[x_index]

                if x_index is None:
                    raise ValueError(f"Find no values, must be: {x_coord}")
                extraction_params["profile_index"] = x_index
            else:
                if not isinstance(y_coord, np.ndarray):
                    y_coord = np.array(y_coord)

                y_index = np.argmin(np.abs(y_coord - coord_val))
                closest_value = y_coord[y_index]

                if y_index is None:
                    raise ValueError(f"Find no values, must be: {y_coord}")
                extraction_params["profile_index"] = y_index

        else:
            raise ValueError(
                "Invalid format for 'profile_position'. Must be None or float position."
            )

        # Return profiles
        if extraction_params["axis"] == "X":
            if isinstance(data, tilupy.read.TemporalResults2D):
                return (
                    tilupy.read.TemporalResults1D(
                        name=data.name,
                        d=data_field[:, extraction_params["profile_index"], :],
                        t=data.t,
                        coords=data.y,
                        coords_name="y",
                        notation=data.notation,
                    ),
                    closest_value,
                )
            else:
                return (
                    tilupy.read.StaticResults1D(
                        name=data.name,
                        d=data_field[:, extraction_params["profile_index"]],
                        coords=data.y,
                        coords_name="y",
                        notation=data.notation,
                    ),
                    closest_value,
                )
        else:
            if isinstance(data, tilupy.read.TemporalResults2D):
                return (
                    tilupy.read.TemporalResults1D(
                        name=data.name,
                        d=data_field[extraction_params["profile_index"], :, :],
                        t=data.t,
                        coords=data.x,
                        coords_name="x",
                        notation=data.notation,
                    ),
                    closest_value,
                )
            else:
                return (
                    tilupy.read.StaticResults1D(
                        name=data.name,
                        d=data_field[extraction_params["profile_index"], :],
                        coords=data.x,
                        coords_name="x",
                        notation=data.notation,
                    ),
                    closest_value,
                )

    elif extraction_mode == "coordinates":
        if "xcoord" not in extraction_params:
            extraction_params["xcoord"] = x_coord[:]
        if "ycoord" not in extraction_params:
            extraction_params["ycoord"] = [0] * len(x_coord)

        # Check errors
        if not isinstance(extraction_params["xcoord"], np.ndarray):
            if isinstance(extraction_params["xcoord"], list):
                extraction_params["xcoord"] = np.array(extraction_params["xcoord"])
            else:
                raise ValueError("Invalid format for 'xcoord'. Must be a numpy array.")
        if not isinstance(extraction_params["ycoord"], np.ndarray):
            if isinstance(extraction_params["ycoord"], list):
                extraction_params["ycoord"] = np.array(extraction_params["ycoord"])
            else:
                raise ValueError("Invalid format for 'ycoord'. Must be a numpy array.")

        if extraction_params["xcoord"].ndim != 1:
            raise ValueError("Invild dimension. 'xcoord' must be a 1d array.")
        if extraction_params["ycoord"].ndim != 1:
            raise ValueError("Invild dimension. 'ycoord' must be a 1d array.")

        if len(extraction_params["xcoord"]) != len(extraction_params["ycoord"]):
            raise ValueError(
                f"'xcoord' and 'ycoord' must have same size: ({len(extraction_params['xcoord'])}, {len(extraction_params['ycoord'])})"
            )

        # Extract index from nearest value of xcoord and ycoord
        x_distances = np.abs(x_coord[None, :] - extraction_params["xcoord"][:, None])
        y_distances = np.abs(y_coord[None, :] - extraction_params["ycoord"][:, None])

        x_indexes = np.argmin(x_distances, axis=1)
        y_indexes = np.argmin(y_distances, axis=1)

        # Compute distance
        x_values = x_coord[x_indexes]
        y_values = y_coord[y_indexes]

        dx = np.diff(x_values)
        dy = np.diff(y_values)

        distance = np.sqrt(dx**2 + dy**2)
        distance = np.concatenate(([0], np.cumsum(distance)))

        # Return profile
        if isinstance(data, tilupy.read.TemporalResults2D):
            return (
                tilupy.read.TemporalResults1D(
                    name=data.name,
                    d=data_field[y_indexes, x_indexes, :],
                    t=data.t,
                    coords=distance,
                    coords_name="d",
                ),
                (x_values, y_values, distance),
            )
        else:
            return (
                tilupy.read.StaticResults1D(
                    name=data.name,
                    d=data_field[y_indexes, x_indexes],
                    coords=distance,
                    coords_name="d",
                ),
                (x_values, y_values, distance),
            )

    elif extraction_mode == "shapefile":
        if "path" not in extraction_params:
            extraction_params["path"] = None
        if "x_origin" not in extraction_params:
            extraction_params["x_origin"] = 0
        if "y_origin" not in extraction_params:
            extraction_params["y_origin"] = y_coord[-1]
        if "x_pixsize" not in extraction_params:
            extraction_params["x_pixsize"] = x_coord[1] - x_coord[0]
        if "y_pixsize" not in extraction_params:
            extraction_params["y_pixsize"] = y_coord[1] - y_coord[0]
        if "step" not in extraction_params:
            extraction_params["step"] = 10

        # Check errors
        if extraction_params["path"] is None:
            raise ValueError("No path to the shape file given.")

        if not isinstance(extraction_params["x_origin"], float) and not isinstance(
            extraction_params["x_origin"], int
        ):
            raise ValueError("'x_origin' must be float.")
        if not isinstance(extraction_params["y_origin"], float) and not isinstance(
            extraction_params["y_origin"], int
        ):
            raise ValueError("'y_origin' must be float.")
        if not isinstance(extraction_params["x_pixsize"], float) and not isinstance(
            extraction_params["x_pixsize"], int
        ):
            raise ValueError("'x_pixsize' must be float.")
        if not isinstance(extraction_params["y_pixsize"], float) and not isinstance(
            extraction_params["y_pixsize"], int
        ):
            raise ValueError("'y_pixsize' must be float.")

        if not isinstance(extraction_params["step"], float) and not isinstance(
            extraction_params["step"], int
        ):
            raise ValueError("'step' must be float.")

        # Import specific module and define extraction function
        from shapely.geometry import LineString, MultiLineString
        from shapely.ops import linemerge
        import geopandas as gpd
        from affine import Affine

        def extract_lines_from_shp_file(shapefile_path):
            """Extract LineString objects from a shapefile."""
            gdf = gpd.read_file(shapefile_path)
            lines = []
            for geom in gdf.geometry:
                if isinstance(geom, LineString):
                    lines.append(geom)
                elif isinstance(geom, MultiLineString):
                    lines.extend(list(geom))
                else:
                    raise TypeError(f"Invalid geometry: {type(geom)}")
            if not lines:
                raise ValueError("No Linestring found.")
            return lines

        lines = extract_lines_from_shp_file(extraction_params["path"])

        # If multiple lines, merge them together
        merged = linemerge(lines)
        if isinstance(merged, LineString):
            profile_line = merged
        else:
            profile_line = LineString([pt for line in merged for pt in line.coords])

        # Construct the affine transformation
        transform = Affine.translation(
            extraction_params["x_origin"], extraction_params["y_origin"]
        ) * Affine.scale(
            extraction_params["x_pixsize"], -extraction_params["y_pixsize"]
        )
        inv = ~transform  # invert : (x, y) -> (row, col)

        distances = np.arange(0, profile_line.length, extraction_params["step"])
        points = [profile_line.interpolate(d) for d in distances]

        # Conversion coordinates -> indexes
        rowcols = [inv * (pt.x, pt.y) for pt in points]
        rowcols = [(int(round(r)), int(round(c))) for c, r in rowcols]

        # Extract values
        all_values = []
        for t in range(len(data.t)):
            values = []
            valid_distances = []
            for d, (r, c) in zip(distances, rowcols):
                if 0 <= r < data_field.shape[0] and 0 <= c < data_field.shape[1]:
                    values.append(data_field[r, c, t])
                    valid_distances.append(d)
            all_values.append(values)

        all_values = np.array(all_values)
        valid_distances = np.array(valid_distances)

        # Return profile
        if isinstance(data, tilupy.read.TemporalResults2D):
            return (
                tilupy.read.TemporalResults1D(
                    name=data.name,
                    d=all_values.T,
                    t=data.t,
                    coords=valid_distances,
                    coords_name="d",
                    notation=data.notation,
                ),
                valid_distances,
            )
        else:
            return (
                tilupy.read.StaticResults1D(
                    name=data.name,
                    d=all_values.T,
                    coords=valid_distances,
                    coords_name="d",
                    notation=data.notation,
                ),
                valid_distances,
            )
    else:
        raise ValueError(
            "Invalid 'extraction_mode': 'axis', 'coordinates' or 'shapefile'."
        )


def format_path_linux(path: str) -> str:
    """
    Change a Windows-type path to a path formatted for Linux. \\ are changed
    to /, and partitions like "C:" are changed to "/mnt/c/"

    Parameters
    ----------
    path : string
        String with the path to be modified.

    Returns
    -------
    path2 : string
        Formatted path.

    """
    if path[1] == ":":
        path2 = "/mnt/{:s}/".format(path[0].lower()) + path[2:]
    else:
        path2 = path
    path2 = path2.replace("\\", "/")
    if " " in path2:
        path2 = '"' + path2 + '"'
    return path2



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os

import numpy as np
import tilupy.read

from tilupy import notations

# Dictionnary with results names lookup table, to match code output names
LOOKUP_NAMES = dict(
    h="rho",
    hvert="rho",
    ux="u",
    uy="ut",
    u="unorm",
    hu="momentum",
    ek="ek",
    vol="vol",
    ep="ep",
    etot="etot",
)
"""Dictionary of correspondance, to match code output names."""

# Classify results
AVAILABLE_OUTPUT = ["h", "hvert", "u", "hu", "hu2", "ux", "uy"]
"""All output available for a shaltop simulation.

Implemented states:

    - h : Flow thickness (normal to the surface)
    - hvert : True vertical flow thickness
    - ux : Velocity component in X direction
    - uy : Velocity component in Y direction

Output computed from other output:

    - u : Norm of the velocity (from ux and uy)
    - hu : Momentum flux (from h and u)
    - hu2 : Convective momentum flux (from h and u)
"""

READ_FROM_FILE_OUTPUT = ["h_max", "u_max", "hu_max", "hu2_int"]
"""Output read from the file.

Read output:

    - h_max : Maximum value of h integrating all time steps
    - u_max : Maximum value of u integrating all time steps
    - hu_max : Maximum value of hu integrating all time steps
    - hu2_int : Integral of values of hu2 for each time steps
"""

FORCES_OUTPUT = []
"""Force-related output variables, including components in x, y, and additional forces.
   
   Generated from:
   
        - faccx, faccy : acceleration force
        - fcurvx, fcurvy : curvature force
        - ffricx, ffricy : friction force
        - fgravx, fgravy : gravitational force
        - finertx, finerty : inertial force
        - fpressionx, fpressiony : pressure force
        
   Additional variables:
    
        - shearx, sheary, shearz : shear stresses
        - normalx, normaly, normalz : normal stresses
        - pbottom : basal pressure
        - pcorrdt : pressure correction (time step)
        - pcorrdiv : pressure correction (divergence)
"""

for force in ["facc", "fcurv", "ffric", "fgrav", "finert", "fpression"]:
    FORCES_OUTPUT.append(force + "x")
    FORCES_OUTPUT.append(force + "y")

FORCES_OUTPUT += [
    "shearx",
    "sheary",
    "shearz",
    "normalx",
    "normaly",
    "normalz",
    "pbottom",
    "pcorrdt",
    "pcorrdiv",
]

INTEGRATED_OUTPUT = ["ek", "ep", "etot"]
"""Integrated energy quantities.

Implemented energy:

    - ek : kinetic energy
    - ep : potential energy
    - etot : total energy
"""


def read_params(file: str) -> dict:
    """Read simulation parameters from file.

    Parameters
    ----------
    file : str
        Path to the parameters file.

    Returns
    -------
    params : dict
        Dictionnary with parameters
    """
    if file is None:
        params = None
    else:
        params = dict()
        with open(file, "r") as f:
            lines = filter(None, (line.rstrip() for line in f))
            for line in lines:
                (key, val) = line.split(" ")
                try:
                    params[key] = float(val)
                except ValueError:
                    params[key] = val.rstrip()
        params["nx"] = int(params["nx"])
        params["ny"] = int(params["ny"])

    return params


def read_file_bin(file: str, nx: int, ny: int) -> np.ndarray:
    """Read shaltop .bin result file.

    Parameters
    ----------
    file : str
        Path to the .bin file.
    nx : int
        Number of cells in the X direction.
    ny : int
        Number of cells in the Y direction.

    Returns
    -------
    numpy.ndarray
        Values in the .bin file.
    """
    data = np.fromfile(file, dtype=np.float32)
    nbim = int(np.size(data) / nx / ny)
    data = np.reshape(data, (nx, ny, nbim), order="F")
    data = np.transpose(np.flip(data, axis=1), (1, 0, 2))
    return data


def read_file_init(file: str, nx: int, ny: int) -> np.ndarray:
    """Read shaltop initial .d files.

    Parameters
    ----------
    file : str
        Path to the .d file.
    nx : int
        Number of cells in the X direction.
    ny : int
        Number of cells in the Y direction.

    Returns
    -------
    numpy.ndarray
        Values in the .d file.
    """
    data = np.loadtxt(file)
    data = np.reshape(data, (nx, ny), order="F")
    data = np.transpose(np.flip(data, axis=1), (1, 0))
    return data


class Results(tilupy.read.Results):
    """Results of shaltop simulations.

    This class is the results class for shaltop. Reading results from shaltop outputs
    are done in this class.

    This class has all the global and quick attributes of the parent class. The quick
    attributes are only computed if needed and can be deleted to clean memory.

    In addition to these attributes, there are those necessary for the operation of
    reading the shaltop results.

    Parameters
    ----------
        folder : str
            Path to the folder containing the simulation files.
        file_params : str
            Name of the simulation parameters file.

    Attributes
    ----------
        _htype : str
            Always "normal".
        _tforces : list
            Times of output.
        _params : dict
            Dictionary of the simulation parameters.
    """

    def __init__(self, folder=None, file_params=None, **varargs):
        super().__init__()
        self._code = "shaltop"

        try:
            if "folder_base" in varargs:
                raise UserWarning(
                    "Variable name has changed: 'folder_base' -> 'folder'"
                )
        except UserWarning as w:
            print(f"[WARNING] {w}")
        if "folder_base" in varargs:
            folder = varargs["folder_base"]

        if folder is None:
            folder = os.getcwd()
        self._folder = folder

        if file_params is None:
            file_params = "params.txt"
        if "." not in file_params:
            file_params = file_params + ".txt"
        file_params = os.path.join(folder, file_params)
        self._params = read_params(file_params)

        # Folder where results are stored
        if "folder_output" not in self._params:
            self._folder_output = os.path.join(self._folder, "data2")
        else:
            self._folder_output = os.path.join(
                self._folder, self._params["folder_output"]
            )

        self._x, self._y = self.get_axes(**self._params)
        self._nx, self._ny = len(self._x), len(self._y)
        self._dx = self._x[1] - self._x[0]
        self._dy = self._y[1] - self._y[0]
        self._zinit = self.get_zinit()

        self._htype = "normal"

        # Get time of outputs
        self._tim = np.loadtxt(os.path.join(self._folder_output, "time_im.d"))
        file_tforces = os.path.join(self._folder_output, "time_forces.d")
        if os.path.isfile(file_tforces):
            self._tforces = np.loadtxt(
                os.path.join(self._folder_output, "time_forces.d")
            )
        else:
            self._tforces = []

    def get_zinit(self, zinit=None) -> np.ndarray:
        """Get zinit, the initial topography.

        Returns
        -------
        numpy.ndarray
            The initial topography.
        """
        path_zinit = os.path.join(self._folder_output, "z.bin")
        if not os.path.isfile(path_zinit) and "file_z_init" in self._params:
            path_zinit = os.path.join(self._folder, self._params["file_z_init"])
            zinit = read_file_init(path_zinit, self._nx, self._ny)
        else:
            zinit = np.squeeze(read_file_bin(path_zinit, self._nx, self._ny))
        return zinit

    def get_axes(self, **varargs) -> tuple[np.ndarray, np.ndarray]:
        """Get X and Y axes.

        varargs : dict
            All parameters needed to compute the axes :
            :data:`x0`, :data:`y0`, :data:`nx`, :data:`ny`, :data:`per`,
            :data:`pery` and :data:`coord_pos`.

        Returns
        -------
        tuple[numpy.ndarray, numpy.ndarray]
            Values of X and Y axes.
        """
        if "x0" in varargs:
            x0 = varargs["x0"]
        else:
            x0 = 0
        if "y0" in varargs:
            y0 = varargs["y0"]
        else:
            y0 = 0

        nx = varargs["nx"]
        ny = varargs["ny"]
        dx = varargs["per"] / nx
        dy = varargs["pery"] / ny
        x = dx * np.arange(nx) + dx / 2
        y = dy * np.arange(ny) + dy / 2

        try:
            coord_pos = varargs["coord_pos"]
        except KeyError:
            coord_pos = "bottom_left"

        if coord_pos == "bottom_left":
            x = x + x0
            y = y + y0
        elif coord_pos == "upper_right":
            x = x + x0 - x[-1]
            y = y + y0 - y[-1]
        elif coord_pos == "upper_left":
            x = x + x0
            y = y + y0 - y[-1]
        elif coord_pos == "lower_right":
            x = x + x0 - x[-1]
            y = y + y0

        return x, y

    def get_u(self) -> np.ndarray:
        """Compute velocity norm from results.

        Returns
        -------
        numpy.ndarray
            Values of flow velocity (norm).
        """
        file = os.path.join(self._folder_output, "u" + ".bin")
        u = read_file_bin(file, self._nx, self._ny)
        file = os.path.join(self._folder_output, "ut" + ".bin")
        ut = read_file_bin(file, self._nx, self._ny)

        if self._costh is None:
            self._costh = self.compute_costh()
        # print(np.shape(self._zinit))
        # print(self._zinit)
        # print(np.shape(self._y))
        # print(np.shape(self._x))

        [Fx, Fy] = np.gradient(self._zinit, np.flip(self._y), self._x)
        u = u * self._costh[:, :, np.newaxis]
        ut = ut * self._costh[:, :, np.newaxis]
        d = np.sqrt(
            u**2 + ut**2 + (Fx[:, :, np.newaxis] * u + Fy[:, :, np.newaxis] * ut) ** 2
        )

        return d

    def _read_from_file(
        self, name: str, operator: str, axis: str = None, **kwargs
    ) -> tilupy.read.StaticResults2D | tilupy.read.TemporalResults0D:
        """Read output from specific files.

        Can access to data in :data:`READ_FROM_FILE_OUTPUT`.

        Parameters
        ----------
        name : str
            Wanted output. Can access to: "u", "hu", "h" and "hu2".
        operator : str
            Wanted operator. Can be "max" for every output except "hu2" or "int" only for "hu2".
        axis : str, optional
            Optional axis. Can be "t". By default None.

        Returns
        -------
        tilupy.read.StaticResults2D | tilupy.read.TemporalResults0D
            Wanted output.
        """
        res = None

        if name in ["u", "hu", "h"]:
            if operator in ["max"] and axis in [None, "t"]:
                file = os.path.join(
                    self._folder_output, LOOKUP_NAMES[name] + operator + ".bin"
                )
                d = np.squeeze(read_file_bin(file, self._nx, self._ny))
                res = tilupy.read.StaticResults2D(
                    "_".join([name, operator]), d, x=self._x, y=self._y, z=self._zinit
                )

        if (name, operator) == ("hu2", "int"):
            array = np.loadtxt(os.path.join(self._folder_output, "ek.d"))
            d = array[:, 1]
            t = array[:, 0]
            res = tilupy.read.TemporalResults0D(name, d, t)

        return res

    def _extract_output(
        self, name: str, **kwargs
    ) -> (
        tilupy.read.TemporalResults2D
        | tilupy.read.TemporalResults0D
        | tilupy.read.AbstractResults
    ):
        """Result extraction for shaltop files.

        Parameters
        ----------
        name : str
            Wanted output. Can access to variables in :data:`AVAILABLE_OUTPUT`, :data:`INTEGRATED_OUTPUT`
            and :data:`FORCES_OUTPUT`.

        Returns
        -------
        tilupy.read.TemporalResults2D | tilupy.read.TemporalResults0D | tilupy.read.AbstractResults
            Wanted output. If no output computed, return an object of :class:`tilupy.read.AbstractResults`.
        """
        # Read thicknesses or velocity components
        d = None
        t = None
        notation = None

        if name in ["h", "ux", "uy", "hvert"]:
            if self._costh is None:
                self._costh = self.compute_costh()

            file = os.path.join(self._folder_output, LOOKUP_NAMES[name] + ".bin")

            d = read_file_bin(file, self._nx, self._ny)

            if name == "hvert":
                d = d / self._costh[:, :, np.newaxis]
            if name in ["ux", "uy"]:
                d = d * self.compute_costh()[:, :, np.newaxis]
            t = self._tim

        if name == "u":
            d = self.get_u()
            t = self._tim

        if name in ["hu", "hu2"]:
            fileh = os.path.join(self._folder_output, "rho.bin")
            h = read_file_bin(fileh, self._nx, self._ny)
            u = self.get_u()
            if name == "hu":
                d = h * u
            elif name == "hu2":
                d = h * u**2
            t = self._tim

        if name in INTEGRATED_OUTPUT:
            array = np.loadtxt(
                os.path.join(self._folder_output, LOOKUP_NAMES[name] + ".d")
            )
            d = array[:, 1]
            if "density" in self._params:
                density = self._params["density"]
            else:
                density = 1
            d = d * density
            t = array[:, 0]

        if name in FORCES_OUTPUT:
            file = os.path.join(self._folder_output, name + ".bin")
            d = read_file_bin(file, self._nx, self._ny)
            t = self._tforces
            notation = notations.Notation(
                name, long_name=name, unit=notations.Unit(Pa=1, kg=-1, m=3), symbol=name
            )

        if d is None:
            file = os.path.join(self._folder_output, name)
            if os.path.isfile(file + ".bin"):
                d = read_file_bin(file + ".bin", self._nx, self._ny)
                t = self._tim
            elif os.path.isfile(file + ".d"):
                d = np.loadtxt(file + ".d")

        if "h_thresh" in kwargs and kwargs["h_thresh"] is not None and d.ndim == 3:
            d = tilupy.read.use_thickness_threshold(self, d, kwargs["h_thresh"])

        if t is None:
            return tilupy.read.AbstractResults(name, d, notation=notation)

        else:
            if d.ndim == 3:
                return tilupy.read.TemporalResults2D(
                    name, d, t, notation=notation, x=self._x, y=self._y, z=self._zinit
                )
            if d.ndim == 1:
                return tilupy.read.TemporalResults0D(name, d, t, notation=notation)
        return None

    # def get_temporal_output(self, name, d=None, t=None, **varargs):
    #     """
    #     Read 2D time dependent simulation results.

    #     Parameters
    #     ----------
    #     name : str
    #         Name of output.
    #     d : ndarray, optional
    #         Data to be read. If None, will be computed.
    #         The default is None.
    #     t : ndarray, optional
    #         Time of results snapshots (1D array), matching last dimension of d.
    #         If None, will be computed. The default is None.
    #     **varargs : TYPE
    #         DESCRIPTION.

    #     """
    #     # Read thicknesses or velocity components
    #     if name in STATES_OUTPUT:
    #         file = os.path.join(
    #             self.folder_output, LOOKUP_NAMES[name] + ".bin"
    #         )
    #         d = read_file_bin(file, self.nx, self.ny)
    #         if name == "hvert":
    #             d = d / self.costh[:, :, np.newaxis]
    #         t = self.tim

    #     # Raed integrated kinetic energy
    #     if name in ["hu2_int"]:
    #         array = np.loadtxt(
    #             os.path.join(self.folder_output, LOOKUP_NAMES[name] + ".d")
    #         )
    #         d = array[:, 1]
    #         t = array[:, 0]

    #     # Compute the velocity
    #     if name == "u":
    #         d = self.get_u()
    #         t = self.tim

    #     if name in ["hu", "hu2"]:
    #         fileh = os.path.join(self.folder_output, "rho.bin")
    #         h = read_file_bin(fileh, self.nx, self.ny)
    #         u = self.get_u()
    #         if name == "hu":
    #             d = h * u
    #         elif name == "hu2":
    #             d = h * u**2
    #         t = self.tim

    #     if (
    #         "h_thresh" in varargs
    #         and varargs["h_thresh"] is not None
    #         and d.ndim == 3
    #     ):
    #         d = tilupy.read.use_thickness_threshold(
    #             self, d, varargs["h_thresh"]
    #         )

    #     return tilupy.read.TemporalResults2D(name, d, t)

    # def get_static_output(self, name, d=None, from_file=True, **varargs):
    #     """
    #     Read 2D time dependent simulation results.

    #     Parameters
    #     ----------
    #     name : str
    #         Name of output.
    #     d : ndarray, optional
    #         Data to be read. Last dimension is for time.
    #         If None, will be computed.
    #         The default is None.
    #     **varargs : TYPE
    #         DESCRIPTION.

    #     """
    #     if name in tilupy.read.COMPUTED_STATIC_DATA_2D:
    #         state, stat = name.split("_")
    #         if stat in ["final", "initial"]:
    #             hh = self.get_temporal_output(state)
    #             if state == "hvert":
    #                 hh.d = hh.d / self.costh[:, :, np.newaxis]
    #             return hh.get_temporal_stat(stat)
    #         if from_file:
    #             file = os.path.join(
    #                 self.folder_output, LOOKUP_NAMES[state] + stat + ".bin"
    #             )
    #             if os.path.isfile(file):
    #                 d = np.squeeze(read_file_bin(file, self.nx, self.ny))
    #             else:
    #                 print(
    #                     file
    #                     + " was not found, "
    #                     + name
    #                     + "_"
    #                     + stat
    #                     + " computed from temporal output."
    #                 )
    #                 from_file = False
    #         if not from_file:
    #             data = self.get_temporal_output(state)
    #             if state == "hvert":
    #                 hh.d = hh.d / self.costh
    #             d = data.get_temporal_stat(stat).d
    #     else:
    #         raise (NotImplementedError())

    #     return tilupy.read.StaticResults(name, d)



# -*- coding: utf-8 -*-

import os

import numpy as np
import tilupy.read

from scipy.interpolate import RegularGridInterpolator


# Classify results
AVAILABLE_OUTPUT = ["h", "hvert", "u", "hu", "hu2"]
"""All output available for a shaltop simulation.

Implemented states:

    - h : Flow thickness (normal to the surface)
    - u : Norm of the velocity (from ux and uy)
    
Output computed from other output:
    
    - hvert : True vertical flow thickness
    - hu : Momentum flux (from h and u)
    - hu2 : Convective momentum flux (from h and u)
"""


class Results(tilupy.read.Results):
    """Results of lave2D simulations.

    This class is the results class for lave2D. Reading results from lave2D outputs
    are done in this class.

    This class has all the global and quick attributes of the parent class. The quick
    attributes are only computed if needed and can be deleted to clean memory.

    In addition to these attributes, there are those necessary for the operation of
    reading the lave2D results.

    Parameters
    ----------
        folder : str
            Path to the folder containing the simulation files.
        name : str
            Simulation/Project name.
        raster : str
            Raster name.
        grid : str, optional
            - If grid=="cell", use cell, output nx and ny sizes are the size of topography raster, minus 1.
            - If grid=="edges" (default), use edges, output will have same dimension as topography raster.

    Attributes
    ----------
        _name : str
            Name of the lave2D project.
        _grid : str
            - If _grid=="cell", use cell, output nx and ny sizes are the size of topography raster, minus 1.
            - If _grid=="edges", use edges, output will have same dimension as topography raster.
        _raster : str
            Path to the raster file.
        _params : dict
            Dictionary storing all simulation parameters.
    """

    def __init__(self, folder, name, raster, grid="edges"):
        super().__init__()
        self._code = "lave2D"

        if folder is None:
            folder = os.getcwd()
        self._folder = folder
        self._folder_output = folder
        # Results are in the same folder than simulation data

        self._name = name

        if not raster.endswith(".asc"):
            raster = raster + ".asc"
        self._raster = raster

        self._x, self._y, self._zinit = tilupy.raster.read_ascii(
            os.path.join(folder, self._raster)
        )
        self._nx, self._ny = len(self._x), len(self._y)
        self._dx = self._x[1] - self._x[0]
        self._dy = self._y[1] - self._y[0]

        # grid=='edges' -> use edges, output will have same dimension as
        # topography raster
        # grid=='cell' -> use cell, output nx and ny sizes are the size of
        # topography raster, minus 1
        self._grid = grid
        if self._grid == "cells":
            fz = RegularGridInterpolator(
                (self._y, self._x),
                self._zinit,
                method="linear",
                bounds_error=False,
                fill_value=None,
            )
            self._x = self._x[1:] - self._dx / 2
            self._y = self._y[1:] - self._dy / 2
            x_mesh, y_mesh = np.meshgrid(self._x, self._y)
            self._zinit = fz((y_mesh, x_mesh))

        self._params = dict()

        self._params["nx"] = len(self._x)
        self._params["ny"] = len(self._y)

        # Rheology
        rheol = np.loadtxt(os.path.join(folder, self._name + ".rhe"))
        self._params["tau/rho"] = rheol[0]
        self._params["K/tau"] = rheol[1]

        # Numerical parameters
        with open(os.path.join(folder, "DONLEDD1.DON"), "r") as fid:
            for line in fid:
                name = line[:34].strip(" ")
                value = line[34:]
                if len(value) == 1:
                    value = int(value)
                else:
                    value = float(value)
                self._params[name] = value

        # Create self._tim
        self._extract_output("X")

    def _extract_output(
        self, name: str, **kwargs
    ) -> tilupy.read.TemporalResults2D | tilupy.read.AbstractResults:
        """Result extraction for lave2D files.

        Parameters
        ----------
        name : str
            Wanted output. Can access to variables in :data:`AVAILABLE_OUTPUT`.

        Returns
        -------
        tilupy.read.TemporalResults2D | tilupy.read.TemporalResults0D | tilupy.read.AbstractResults
            Wanted output. If no output computed, return an object of :class:`tilupy.read.AbstractResults`.
        """
        d = None
        t = None
        notation = None

        # read initial mass
        h_init = np.loadtxt(os.path.join(self._folder, self._name + ".cin"), skiprows=1)

        # Read results
        file_res = os.path.join(self._folder, self._name + ".asc")
        n_times = int(self._params["tmax"] / self._params["dtsorties"])
        tim = [0]

        if self._grid == "edges":
            ny_out = self._params["ny"] - 1
            x_out = self._x[1:] - self._dx / 2
            y_out = self._y[1:] - self._dy / 2
            x_mesh, y_mesh = np.meshgrid(self._x, self._y)
            h_init = h_init.reshape(self._params["ny"] - 1, self._params["nx"] - 1)
            fh = RegularGridInterpolator(
                (y_out, x_out),
                h_init,
                method="linear",
                bounds_error=False,
                fill_value=None,
            )
            h_init = fh((y_mesh, x_mesh))
        else:
            ny_out = self._params["ny"]
            h_init = h_init.reshape(self._params["ny"], self._params["nx"])

        x_mesh_out, y_mesh_out = np.meshgrid(
            self._x[1:] - self._dx / 2, self._y[1:] - self._dy / 2
        )

        h = np.zeros((self._params["ny"], self._params["nx"], n_times + 1))
        h[:, :, 0] = np.flip(h_init, axis=0)
        u = np.zeros((self._params["ny"], self._params["nx"], n_times + 1))

        with open(file_res, "r") as fid:
            lines = fid.readlines()
            n_lines = 2 * (ny_out) + 8  # Number of lines per time step
            for i_time in range(n_times):
                i_start = n_lines * i_time
                i_start_h = i_start + 7
                i_stop_h = i_start + ny_out + 7
                i_start_u = i_start + ny_out + 8
                i_stop_u = i_start + 2 * ny_out + 8
                tim.append(float(lines[i_start]))
                h_out = np.loadtxt(lines[i_start_h:i_stop_h])
                u_out = np.loadtxt(lines[i_start_u:i_stop_u])

                if self._grid == "edges":
                    fh = RegularGridInterpolator(
                        (y_out, x_out),
                        h_out,
                        method="linear",
                        bounds_error=False,
                        fill_value=None,
                    )
                    h[:, :, i_time + 1] = fh((y_mesh, x_mesh))

                    fu = RegularGridInterpolator(
                        (y_out, x_out),
                        u_out,
                        method="linear",
                        bounds_error=False,
                        fill_value=None,
                    )
                    u[:, :, i_time + 1] = fu((y_mesh, x_mesh))
                else:
                    h[:, :, i_time + 1] = h_out
                    u[:, :, i_time + 1] = u_out

        if self._tim is None:
            self._tim = np.array(tim)

        extracted_outputs = {
            "h": h[:],
            "u": u[:],
        }

        if name in ["h", "u"]:
            d = extracted_outputs[name]
            t = self._tim

        if name == "hu":
            d = extracted_outputs["h"] * extracted_outputs["u"]
            t = self._tim

        elif name == "hu2":
            d = extracted_outputs["h"] * extracted_outputs["u"] * extracted_outputs["u"]
            t = self._tim

        elif name == "hvert":
            if self._costh is None:
                self._costh = self.compute_costh()
            d = extracted_outputs["h"] / self._costh[:, :, np.newaxis]
            t = self._tim

        # if name == "ek":
        #     if self._costh is None:
        #             self._costh = self.compute_costh()

        #     d = []
        #     for i in range(len(self._tim)):
        #         d.append(np.sum((available_outputs['h'][:, :, i]
        #                          * available_outputs['u'][:, :, i]
        #                          * available_outputs['u'][:, :, i])
        #                         / self._costh[:, :]))
        #     d = np.array(d)
        #     t = self._tim

        if t is None:
            return tilupy.read.AbstractResults(name, d, notation=notation)

        else:
            if d.ndim == 3:
                return tilupy.read.TemporalResults2D(
                    name, d, t, notation=notation, x=self._x, y=self._y, z=self._zinit
                )
        return None

    def _read_from_file(self, *args, **kwargs):
        """Not useful"""
        return "No _read_from_file for Lave2D."



import pytest
import numpy as np
import tilupy.utils


@pytest.fixture
def axis():
    x = np.linspace(0, 1, 5)
    y = np.flip(np.linspace(0, 2, 10))
    return dict(x=x, y=y)


@pytest.mark.parametrize(
    "ax, ay",
    [(0.5, 0.5), (0.5, -1), (-1, 0.5)],
)
def test_normal_vector(axis, ax, ay):
    z = ax * axis["x"][np.newaxis, :] + ay * axis["y"][:, np.newaxis]
    nn = np.sqrt(1 + ax**2 + ay**2)
    nx = -ax / nn
    ny = -ay / nn
    nz = 1 / nn
    Nx, Ny, Nz = tilupy.utils.normal_vector(axis["x"], axis["y"], z)
    assert Nx.shape == Ny.shape == Nz.shape == (len(axis["y"]), len(axis["x"]))
    assert np.allclose(Nx, nx)
    assert np.allclose(Ny, ny)
    assert np.allclose(Nz, nz)



import pytest
import os

import tilupy.models.lave2D.read as laread
import tilupy.read as tiread


@pytest.fixture
def simu_data():
    res = dict(
        folder=os.path.join(os.path.dirname(__file__), "data", "gray99", "lave2D"),
        simu_name="projects",
        raster_name="toposimu",
    )
    return res


@pytest.mark.parametrize(
    "args, expected",
    [
        ("h", (3, tiread.TemporalResults2D)),
        ("hvert", (3, tiread.TemporalResults2D)),
        ("u", (3, tiread.TemporalResults2D)),
        ("hu", (3, tiread.TemporalResults2D)),
        ("hu2", (3, tiread.TemporalResults2D)),
    ],
)
def test_extract_output(simu_data, args, expected):
    res = laread.Results(
        folder=simu_data["folder"],
        name=simu_data["simu_name"],
        raster=simu_data["raster_name"],
    )
    output = res._extract_output(args)

    assert output.d.ndim == expected[0]
    assert isinstance(output, expected[1])


def test_extract_invalid_output(simu_data):
    res = laread.Results(
        folder=simu_data["folder"],
        name=simu_data["simu_name"],
        raster=simu_data["raster_name"],
    )
    output = res._extract_output("xx")

    assert isinstance(output, tiread.AbstractResults)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Jun  2 17:57:31 2021

@author: peruzzetto
"""

import importlib
import os
import string
import numpy as np
import matplotlib.pyplot as plt
import mpl_toolkits.axes_grid1 as mplt

# import tilupy.dem
import tilupy.plot as plt_fn
import tilupy.notations
import tilupy

from tilupy.notations import LABELS
from tilupy.read import STATIC_DATA_2D


def compare_spatial_results(
    results,
    name,
    stat,
    import_kwargs=None,
    x=None,
    y=None,
    zinit=None,
    nrows_ncols=None,
    figsize=None,
    folder_out=None,
    file_prefix=None,
    fmt="png",
    dpi=150,
    cbar_location="bottom",
    cbar_size="5%",
    cmap_intervals=None,
    cmap="inferno",
    topo_kwargs=None,
    title_loc="top right",
    title_x_offset=0.02,
    title_y_offset=0.02,
):
    """
    Compare spatiale results of different simulations

    Parameters
    ----------
    results : TYPE
        DESCRIPTION.
    name : TYPE
        DESCRIPTION.
    import_kwargs : TYPE, optional
        DESCRIPTION. The default is None.
    x : TYPE, optional
        DESCRIPTION. The default is None.
    y : TYPE, optional
        DESCRIPTION. The default is None.
    zinit : TYPE, optional
        DESCRIPTION. The default is None.
    figsize : TYPE, optional
        DESCRIPTION. The default is None.
    folder_out : TYPE, optional
        DESCRIPTION. The default is None.
    cbar_location : TYPE, optional
        DESCRIPTION. The default is 'bottom'.
    cbar_size : TYPE, optional
        DESCRIPTION. The default is "5%".
    cmap_intervals : TYPE, optional
        DESCRIPTION. The default is None.
    cmap : TYPE, optional
        DESCRIPTION. The default is 'inferno'.
    topo_kwargs : TYPE, optional
        DESCRIPTION. The default is None.

    Returns
    -------
    None.

    """
    if x is None or y is None:
        results[0].set_axes()
        x = results[0].x
        y = results[0].y

    if zinit is None:
        results[0].set_zinit()
        zinit = results[0].zinit

    topo_kwargs = {} if topo_kwargs is None else topo_kwargs
    import_kwargs = {} if import_kwargs is None else import_kwargs
    nplots = len(results)
    if nrows_ncols is None:
        nrows_ncols = (1, nplots)

    fig = plt.figure(figsize=figsize)
    axes = mplt.AxesGrid(
        fig,
        111,
        nrows_ncols=nrows_ncols,
        axes_pad=0.1,
        share_all=True,
        aspect="True",
        label_mode="L",
        cbar_location=cbar_location,
        cbar_mode="single",
        cbar_pad="12%",
        cbar_size=cbar_size,
    )

    # Titles
    alignments = title_loc.split(" ")
    ha = alignments[1]
    va = alignments[0]
    if va == "top":
        xt = 1 - title_x_offset
    elif va == "center":
        xt = 0.5
    elif va == "bottom":
        xt = title_x_offset
    if ha == "right":
        yt = 1 - title_y_offset
    elif ha == "center":
        yt = 0.5
    elif ha == "bottom":
        yt = title_y_offset

    for i, result in enumerate(results):
        data = result.get_static_output(name, stat, **import_kwargs)
        plt_fn.plot_data_on_topo(
            x,
            y,
            zinit,
            data.d,
            axe=axes[i],
            plot_colorbar=False,
            colorbar_kwargs={"extend": "max"},
            cmap_intervals=cmap_intervals,
            cmap=cmap,
            topo_kwargs=topo_kwargs,
        )
        letter = "({:s})".format(string.ascii_lowercase[i])
        axes[i].text(
            xt,
            yt,
            "{:s} {:s}".format(letter, result.code),
            ha=ha,
            va=va,
            fontsize=11,
            zorder=100,
            bbox=dict(boxstyle="round", fc="w", ec="k", pad=0.2),
            transform=axes[i].transAxes,
        )
        axes[i].grid(False)

    # Colorbar
    axe_cc = axes.cbar_axes[0]
    if cbar_location in ["right", "left"]:
        orientation = "vertical"
    else:
        orientation = "horizontal"
    fig.colorbar(
        axes[0].images[-1],
        cax=axe_cc,
        ax=axes[0],
        orientation=orientation,
        extend="max",
    )
    if orientation == "vertical":
        axe_cc.set_ylabel(LABELS[name + stat])
    else:
        axe_cc.set_xlabel(LABELS[name + stat])

    fig.tight_layout()

    if folder_out is not None:
        if file_prefix is None:
            file = name + stat + "." + fmt
        else:
            file = file_prefix + "_" + name + stat + "." + fmt
        fig.savefig(os.path.join(folder_out, file), dpi=dpi)


def compare_simus(
    codes,
    topo_name,
    law,
    rheol_params,
    output_name,
    stat_name,
    subfolder_topo="",
    folder_benchmark=None,
    folder_out=None,
    **kwargs,
):
    """Compare simulations with different codes."""

    modules = dict()
    for c in codes:
        modules[c] = importlib.import_module("tilupy.models." + c + ".read")

    if folder_benchmark is None:
        folder_benchmark = os.getcwd()

    folder_topo = os.path.join(folder_benchmark, topo_name)

    txt_params = tilupy.notations.make_rheol_string(rheol_params, law)

    if folder_out is None:
        folder_out = os.path.join(folder_topo, subfolder_topo, "comparison", law)
        os.makedirs(folder_out, exist_ok=True)

    file = np.os.path.join(folder_topo, "topo.asc")
    x, y, zinit, dx = tilupy.dem.read_ascii(file)

    simus = []
    for code in codes:
        folder_simus = os.path.join(folder_topo, subfolder_topo, code, law)
        simus.append(modules[code].Results(txt_params, folder_simus))

    if output_name + stat_name in STATIC_DATA_2D:
        compare_spatial_results(
            simus,
            output_name,
            stat_name,
            x=x,
            y=y,
            zinit=zinit,
            folder_out=folder_out,
            file_prefix=txt_params,
            **kwargs,
        )



import pandas as pd
import tilupy.download_data


def test_import_shaltop_mus_calibrated():
    df, _ = tilupy.download_data.import_shaltop_mus_calibrated()
    assert isinstance(df, pd.DataFrame)
    assert "mus" in df.columns
